"""
Module to provide Telemetry class as public class to be accessed

classes
    - Telemetry

"""
from collections import defaultdict
from datetime import datetime
from typing import List

from errors import ErrorCode

from .settings import exceptions
from .settings.data_class import ProcessType, TelemetryCounter
from .settings.process_type import ProcessTypes
from .settings.settings import DEFAULT_TRAFIC_LIGHT_COLOR, \
    TRAFIC_LIGHT_COLOR_ORANGE, TRAFIC_LIGHT_COLOR_RED
from .storage.generic import AbstractTelemetryStorage
from .storage.memory import TelemetryInMemoryStorage
from .validators.dict_validator import DictValidator

# default telemetry field names
BASE_COUNT_KEY = "base_counter"
ERRORS_KEY = "errors"
FAIL_COUNT_KEY = "fail_counter"
SOURCE_NAME_KEY = "source_name"
CATEGORY_KEY = "category"
SUB_CATEGORY_KEY = "sub_category"
PROCESS_TYPE_KEY = "process_type"
START_TIME = "start_date_time"
RUN_TIME = "run_time_in_seconds"
TRAFFIC_LIGHT_KEY = "traffic_light"


# decorator method to check status telemetry object
def _raise_exception_if_telemetry_closed(method):
    """
    Decorator method to check if telemetry object is closed.
    If so an exception is raised.

    Decorator method to be used for methods that are only allowed
    when telemetry object not yet closed.
    """

    def wrapper(self, *args, **kwargs):
        """
        Wrapper to check if run_time has been set
        If so the telemetry object is closed
        """
        if self.telemetry.get(RUN_TIME):
            raise exceptions.TelemetryObjectAlreadyClosed()

        return method(self, *args, **kwargs)

    return wrapper


class Telemetry:
    """Class to manage the telemetry data object of a data pipeline process.

    This class can be used to measure and store indicators of a data process.
    When the dataprocess is finished this Telemetry object can be persisted in
    a database provide via a storage class.

    args:
        - source_name (str):
            Process name for which telemetry is made for example: `GET-WEATHER`
            Process names can be choosen freely and should ne unique for a
            single process as they are used to collect all telemetry data for a
            specific process.
        - process_type (ProcessType):
            Process Type for which the telemetry is created. For example
            `CREATE_DATA_FROM_API`. Process type are predefined objects and
            made available via class ProcessTypes. Each process type defines
            a number of sub processes like for example `RETRIEVE_RAW_DATA`. For
            these subprocess specific telemetry data can be stored in the
            telemetry instance
        - telemetry_rules (dict)
            rules how to evaluate the data object for specific sub process
            see DictValidator class for more details on how to define rules
        - storage_class (AbstractTelemetryStorage)
            class to define storage method for Telemetry object

    public class methods:
        - add_process_types: add custom process types

    properties:
        - category
        - sub_category
        - source_name
        - process_type
        - sub_process_types

    public methods:
        - get (as normal get method for dict)
        - save_and_close
        - add
        - increase_sub_process_base_count
        - increase_sub_process_fail_count
        - increase_sub_process_custom_count
        - increase_custom_count
    """

    _available_process_types = ProcessTypes

    def __init__(
        self,
        category: str,
        sub_category: str,
        source_name: str,
        process_type: ProcessType,
        telemetry_rules: dict = None,
        storage_class: AbstractTelemetryStorage = None,
    ):
        self._set_process_type(process_type)
        self._telemetry = defaultdict(int)
        self._telemetry.update({
            CATEGORY_KEY: category,
            SUB_CATEGORY_KEY: sub_category,
            SOURCE_NAME_KEY: source_name,
            PROCESS_TYPE_KEY: self._process_type.name,
            START_TIME: datetime.now(),
            TRAFFIC_LIGHT_KEY: DEFAULT_TRAFIC_LIGHT_COLOR
        })
        self._telemetry_rules = telemetry_rules or {}
        self._storage_class = self._get_storage_class(storage_class)

    @classmethod
    def add_process_type(cls, process_type_key: str, process_type: ProcessType) -> None:
        """
        Add a custom process type to the available process types to the
        already registered process types.

        Args:
            process_type (ProcessType): Process type that needs to be added
            process_type_key (str): key that should be used when calling this
                                    ProcessType
        """
        if not isinstance(process_type, ProcessType):
            raise exceptions.ProcessTypeMustBeOfClassProcessType()
        cls._available_process_types.register_process_type(
            process_type_key, process_type
        )

    @staticmethod
    def _get_storage_class(storage_class: AbstractTelemetryStorage) -> None:
        """
        validates and sets the storage class. If no storage class provided
        it is defaulted to TelemetryInMemoryStorage

        Args:
            storage_class (AbstractTelemetryStorage):
                Storage class, child class from AbstractTelemetryStorage

        Raises:
            StorageClassOfIncorrectType
        """
        # set default in memory storage if no storage class provided
        if not storage_class:
            storage_class = TelemetryInMemoryStorage

        if not issubclass(storage_class, AbstractTelemetryStorage):
            raise exceptions.StorageClassOfIncorrectType(
                storage_class.__class__.__name__
            )

        return storage_class()

    @property
    def storage_class(self):
        return self._storage_class

    @storage_class.setter
    def storage_class(self, storage_class):
        self._storage_class = self._get_storage_class(storage_class)

    @property
    def source_name(self) -> str:
        """Source_name property."""
        return self._telemetry.get(SOURCE_NAME_KEY)

    @property
    def traffic_light(self) -> str:
        """Traffic_light property."""
        return self._telemetry.get(TRAFFIC_LIGHT_KEY)

    @property
    def telemetry(self) -> dict:
        """Telemetry property."""
        return self._telemetry

    @property
    def sub_process_types(self) -> list:
        """Returns a of subprocess types allowed for the Telemetry instance."""
        return self._process_type.sub_processes

    def set_orange_traffic_light(self) -> None:
        """Sets traffic light attribute to orange."""
        self._telemetry[TRAFFIC_LIGHT_KEY] = TRAFIC_LIGHT_COLOR_ORANGE

    def set_red_traffic_light(self) -> None:
        """Sets traffic light attribute to red."""
        self._telemetry[TRAFFIC_LIGHT_KEY] = TRAFIC_LIGHT_COLOR_RED

    def get(self, field_name: str) -> dict:
        """Retrieve a field from the telemetry object.

        Implements the get method on the Telemetry object as if it is a dict

        Returns:
            dict: [description]
        """
        return self._telemetry.get(field_name)

    def save_and_close(self) -> dict:
        """
        Closes and stores the telemetry instance and returns the
        telemetry value.

        Returns:
            dict: telemetry result
        """
        if self._telemetry.get(RUN_TIME):
            raise exceptions.TelemetryObjectAlreadyClosed()
        self._set_runtime()
        self._storage_class.store_telemetry(self._telemetry)

        return self._telemetry

    def _set_runtime(self) -> None:
        """Method to calculate and set the runtime seconds (in str)."""
        run_time = datetime.now() - self._telemetry.get(START_TIME)
        self._telemetry[RUN_TIME] = str(run_time.total_seconds())

    @_raise_exception_if_telemetry_closed
    def add(self, sub_process: str, data: dict, errors: List[ErrorCode]) -> None:
        """
        Add data validation errors and/or errors from data process to a
        telemetry sub process.
        Data validation errors are retrieved from data validation defined in
        telemetry rules.

        Method does not update BASE_COUNT.

        :param sub_process: applicable subprocess for the errors
        :type sub_process: str
        :param data: data generated by subproces for which telemetry
                     needs to be generated
        :param type: dict
        :param errors: errors from data process to be added
        :type errors: list of Errorcodes
        :returns: None
        """
        if not self._telemetry.get(sub_process):
            self._initialize_sub_process(sub_process)
        validation_errors = self._validate_data(sub_process, data)
        # add data validation errors
        self._add_errors(sub_process, validation_errors)

        # add data process errors
        if errors:
            self._add_errors(sub_process, errors)

    def _validate_data(self, sub_process: str, data: dict) -> List[ErrorCode]:
        """Validates the data provided by a subprocess.

        Data is validated based upon telemetry rules for the subprocces

        :param sub_process: applicable subprocess
        :typesub_process: str
        :param data: data provided by subprocess
        :type data: dict
        :returns: list of Errorcodes
        """
        telemetry_rules = self._telemetry_rules.get(sub_process, {})
        return DictValidator().validate(data, telemetry_rules)

    def _add_errors(self, sub_process: str, errors: list) -> None:
        """Adds the error to a telemetry sub process.

        :param sub_process: applicable subprocess for the errors
        :typesub_process: str
        :param errors: errors to be processed
        :type errors: list of Errorcodes
        :returns: None
        """
        for error in errors:
            self.increase_sub_process_error_count(error=error, sub_process=sub_process)

    @_raise_exception_if_telemetry_closed
    def increase_sub_process_base_count(
        self, sub_process: str, increment: int = 1
    ) -> None:
        """
        Increases the base count for a subprocess.

        Args:
            sub_process (str): name of subprocess
        """
        if not self._telemetry.get(sub_process):
            self._initialize_sub_process(sub_process)

        self._telemetry[sub_process][BASE_COUNT_KEY] += increment

    @_raise_exception_if_telemetry_closed
    def increase_sub_process_fail_count(
        self, sub_process: str, increment: int = 1
    ) -> None:
        """Increases the fail count for a subprocess

        Args:
            sub_process (str): name of subprocess
        Raises:
            BaseCountForSubProcessNotAdded: if subprocess has not yet been
                                            created
        """
        if not self._telemetry.get(sub_process):
            raise exceptions.BaseCountForSubProcessNotAdded(sub_process)

        self._telemetry[sub_process][FAIL_COUNT_KEY] += increment

    @_raise_exception_if_telemetry_closed
    def increase_sub_process_custom_count(
        self, custom_counter: str, sub_process: str, increment: int = 1
    ) -> None:
        """
        Increases a custom counter for a subprocess.

        Args:
            sub_process (str): name of subprocess
            custom_counter (str): name of custom counter
            increment (int): increment for the counter (default = 1)

        Raises:
            BaseCountForSubProcessNotAdded: if subprocess has not yet been
                                            created
        """
        if not self._telemetry.get(sub_process):
            raise exceptions.BaseCountForSubProcessNotAdded(sub_process)

        self._telemetry[sub_process][custom_counter] += increment

    @_raise_exception_if_telemetry_closed
    def increase_sub_process_error_count(
        self, sub_process: str, error: ErrorCode, increment: int = 1
    ) -> None:
        """
        Increases a error counter for a subprocess.

        Args:
            sub_process (str): name of subprocess
            error_code (str): error code from ErrorCode object
            increment (int): increment for the counter

        Raises:
            BaseCountForSubProcessNotAdded: if subprocess has not yet been
                                            created
        """
        if not self._telemetry.get(sub_process):
            raise exceptions.BaseCountForSubProcessNotAdded(sub_process)

        self._telemetry[sub_process][ERRORS_KEY][error.code] += increment

    @_raise_exception_if_telemetry_closed
    def increase_custom_count(self, custom_counter: str, increment: int = 1) -> None:
        """
        Increases a custom counter in the top level of the telemetry object.
        When then custom counter does not yet exist is created.

        Args:
            custom_counter (str): name of custom counter
            increment (int): increment for the counter (Default = 1)
        """
        self._telemetry[custom_counter] += increment

    def add_telemetry_counter(
        self, telemetry_counter: TelemetryCounter, increment: int = None
    ) -> None:
        """
        Method to process a TelemetryCounter object with predefined counters.

        Args:
            telemetry_counter (TelemetryCounter): [description]
            increment (int, None):
                when not None overules the increment setting from
                TelemetryCounter (which defaults to 1)
        """
        telemetry_counter.validate_sub_process()
        sub_process = telemetry_counter.sub_process
        counter_name = telemetry_counter.counter_name
        increment = increment or telemetry_counter.increment
        if telemetry_counter.error:
            self.increase_sub_process_error_count(
                sub_process=sub_process,
                error=telemetry_counter.error,
                increment=increment,
            )
        if not telemetry_counter.error:
            self.increase_sub_process_custom_count(
                sub_process=sub_process,
                custom_counter=counter_name,
                increment=increment,
            )

    def _initialize_sub_process(self, sub_process: str) -> None:
        """sets the initial count object for a sub_process

        Args:
            sub_process (str): sub process name
        """
        if sub_process not in self._process_type.sub_processes:
            raise exceptions.InvalidSubProcess(sub_process, self._process_type)

        self._telemetry[sub_process] = defaultdict(int)
        self._telemetry[sub_process].update({
            BASE_COUNT_KEY: 0,
            FAIL_COUNT_KEY: 0,
            ERRORS_KEY: defaultdict(int)
        })

    def _set_process_type(self, process_type: ProcessType) -> None:
        """Sets the process type for the telemetrty instance."""
        if not isinstance(process_type, ProcessType):
            raise exceptions.ProcessTypeMustBeOfClassProcessType
        if not self._available_process_types.is_registered(process_type):
            raise exceptions.ProcessTypeNotRegistered(process_type)
        self._process_type = process_type



"""
Module to test telemetry main class for pipeline telemetry module.
"""
from datetime import datetime

import pytest
from test_data import DEFAULT_TELEMETRY_PARAMS

from pipeline_telemetry.main import FAIL_COUNT_KEY, Telemetry
from pipeline_telemetry.settings import DEFAULT_TRAFIC_LIGHT_COLOR, \
    TRAFIC_LIGHT_COLOR_ORANGE, TRAFIC_LIGHT_COLOR_RED, exceptions, settings
from pipeline_telemetry.settings.data_class import ProcessType
from pipeline_telemetry.storage.generic import AbstractTelemetryStorage
from pipeline_telemetry.storage.memory import TelemetryInMemoryStorage

# pylint: disable=protected-access


def test_telemetry_exists():
    """check that Telemetry class exists"""
    assert Telemetry


def test_telemetry_instance_creation():
    """check that Telemetry instance can be created using default_params"""
    assert Telemetry(**DEFAULT_TELEMETRY_PARAMS).source_name == "load_weather_data"


def test_telemetry_instance_creation_raises_excption():
    """
    Check that Telemetry instance creation raises exception with non registered
    ProcessType.
    """
    process_type = ProcessType(process_type="not_registered", subtypes=["test"])
    telemetry_params = {
        "category": "WEATHER",
        "sub_category": "DAILY_PREDICTIONS",
        "source_name": "load_weather_data",
        "process_type": process_type,
    }
    with pytest.raises(exceptions.ProcessTypeNotRegistered):
        Telemetry(**telemetry_params)


def test_telemetry_instance_has_telemetry_property():
    """check that Telemetry instance has a telemetry property"""
    assert Telemetry(**DEFAULT_TELEMETRY_PARAMS).telemetry


def test_telemetry_instance_has_sub_process_types_property():
    """
    check that Telemetry instance has a sub_process_types property that return
    a dict with all sub_process_types
    """
    assert isinstance(Telemetry(**DEFAULT_TELEMETRY_PARAMS).sub_process_types, list)
    assert (
        Telemetry(**DEFAULT_TELEMETRY_PARAMS).sub_process_types
        == settings.DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES
    )


def test_telemetry_instance_raises_exception_with_invalid_process_type():
    """
    check that Telemetry instance creation raises exception when invalid
    process_type is used
    """
    telemetry_params = DEFAULT_TELEMETRY_PARAMS | {"process_type": "invalid"}
    with pytest.raises(exceptions.ProcessTypeMustBeOfClassProcessType):
        Telemetry(**telemetry_params)


def test_telemetry_instance_sets_start_time():
    """
    check that Telemetry instance has a start_date_time
    """
    telemetry = Telemetry(**DEFAULT_TELEMETRY_PARAMS).telemetry
    assert isinstance(telemetry.get("start_date_time"), datetime)


def test_increase_new_sub_process_base_count_to_telemetry():
    """
    Check that a new sub process base count can be added to a Telemetry
    instance.
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    assert "RETRIEVE_RAW_DATA" in telemetry_inst.telemetry
    assert telemetry_inst.get("RETRIEVE_RAW_DATA")["base_counter"] == 1


def test_increase_non_exsiting_sub_process_raises_exception():
    """
    Check that a add a based count for a non existing sub process raises
    an excpetion
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    with pytest.raises(exceptions.InvalidSubProcess):
        telemetry_inst.increase_sub_process_base_count("Non exsiting")


def test_increase_sub_process_base_count_to_telemetry():
    """
    Check that an existing sub process base count can be inreased.
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    assert telemetry_inst.get("RETRIEVE_RAW_DATA")["base_counter"] == 1
    telemetry_inst.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    assert telemetry_inst.get("RETRIEVE_RAW_DATA")["base_counter"] == 2


def test_increase_sub_process_count_not_allowed_with_closed_telemetry():
    """
    check that a sub process base count can be added to th eTelemetry instance
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    telemetry_inst.save_and_close()
    with pytest.raises(exceptions.TelemetryObjectAlreadyClosed):
        telemetry_inst.increase_sub_process_base_count("RETRIEVE_RAW_DATA")


def test_can_not_close_closed_telemetry_instance():
    """
    check that a closed telemetry instance can not be closed again
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.save_and_close()
    with pytest.raises(exceptions.TelemetryObjectAlreadyClosed):
        telemetry_inst.save_and_close()


def test_increase_sub_process_fail_not_allowed_with_closed_telemetry():
    """
    check that a sub process base count can be added to th eTelemetry instance
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    telemetry_inst.save_and_close()
    with pytest.raises(exceptions.TelemetryObjectAlreadyClosed):
        telemetry_inst.increase_sub_process_fail_count("RETRIEVE_RAW_DATA")


def test_sub_process_custom_count_not_allowed_with_closed_telemetry():
    """
    check that a sub process custom count method raises an excpetion
    when telementry object has been closed
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    telemetry_inst.save_and_close()
    with pytest.raises(exceptions.TelemetryObjectAlreadyClosed):
        telemetry_inst.increase_sub_process_custom_count(
            sub_process="RETRIEVE_RAW_DATA", custom_counter="test_counter"
        )


def test_custom_count_not_allowed_with_closed_telemetry():
    """
    check that a custom count method raises an excpetion
    when telementry object has been closed
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.save_and_close()
    with pytest.raises(exceptions.TelemetryObjectAlreadyClosed):
        telemetry_inst.increase_custom_count(custom_counter="test_counter")


def test_add_sub_process_fail_count_to_telemetry():
    """
    check that a sub process fail count can be increased
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    telemetry_inst.increase_sub_process_fail_count("RETRIEVE_RAW_DATA")
    assert telemetry_inst.telemetry["RETRIEVE_RAW_DATA"][FAIL_COUNT_KEY] == 1


def test_add_sub_process_fail_count_raises_excpetion():
    """
    check that a sub process fail count raises an excpetion when sub_process
    has not yet been initialized by increase_sub_process_base_count(sub_process)
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    with pytest.raises(exceptions.BaseCountForSubProcessNotAdded):
        telemetry_inst.increase_sub_process_fail_count("RETRIEVE_RAW_DATA")


def test_add_custom_counter_to_telemetry():
    """
    check that a custom counter can be added to the telemetry instance
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_custom_count("network_error")
    assert telemetry_inst.telemetry["network_error"] == 1


def test_add_sub_process_custom_counter_to_telemetry():
    """
    Check that a custom counter can be added to a subprocess.
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    telemetry_inst.increase_sub_process_custom_count(
        custom_counter="network_error", sub_process="RETRIEVE_RAW_DATA"
    )
    assert telemetry_inst.get("RETRIEVE_RAW_DATA")["network_error"] == 1


def test_increase_existing_sub_process_custom_counter_to_telemetry():
    """
    Check that a existing custom counter can be icnrease
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    for _ in [1, 2]:
        telemetry_inst.increase_sub_process_custom_count(
            custom_counter="network_error", sub_process="RETRIEVE_RAW_DATA"
        )

    assert telemetry_inst.get("RETRIEVE_RAW_DATA")["network_error"] == 2


def test_increase_custom_counter_to_telemetry():
    """
    check that a custom counter once created can be increased with any value
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_custom_count("network_error")
    telemetry_inst.increase_custom_count("network_error", 2)
    assert telemetry_inst.telemetry["network_error"] == 3


def test_increase_sub_process_custom_count_raises_excpetion():
    """
    check that increase_sub_process_custom_count method raises an excpetion
    when sub_process has not yet been initialized
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    with pytest.raises(exceptions.BaseCountForSubProcessNotAdded):
        telemetry_inst.increase_sub_process_custom_count(
            sub_process="RETRIEVE_RAW_DATA", custom_counter="test_counter"
        )


def test_increase_sub_process_custom_count_with_two():
    """
    check that increase_sub_process_custom_count method creates counter
    with initial value 2
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    telemetry_inst.increase_sub_process_custom_count(
        sub_process="RETRIEVE_RAW_DATA", custom_counter="test_counter", increment=2
    )
    assert telemetry_inst.telemetry["RETRIEVE_RAW_DATA"]["test_counter"] == 2


def test_close_telemetry_instance_sets_run_time(mocker):
    """
    check that closing the telemetry sets the run_time
    """
    mocker.patch(
        (
            "pipeline_telemetry.storage.memory."
            "TelemetryInMemoryStorage.store_telemetry"
        ),
        return_value=None,
    )
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    telemetry_result = telemetry_inst.save_and_close()
    assert float(telemetry_result.get("run_time_in_seconds")) > 0


def test_close_telemetry_instance_calls_store_telemetry(mocker):
    """
    check that closing the telemetry sets the run_time
    """
    mocker.patch(
        (
            "pipeline_telemetry.storage.memory."
            "TelemetryInMemoryStorage.store_telemetry"
        ),
        return_value=None,
    )
    _store_telemetry_spy = mocker.spy(TelemetryInMemoryStorage, "store_telemetry")
    Telemetry(**DEFAULT_TELEMETRY_PARAMS).save_and_close()
    assert _store_telemetry_spy.called


def test_get_storage_class_returns_in_memory_storage_by_default():
    """
    check that _get_storage_class method returns TelemetryInMemoryStorage
    instance if no storage class is provided
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    assert isinstance(telemetry_inst._get_storage_class(None), TelemetryInMemoryStorage)


def test_get_storage_class_raises_exception_if_invalid_class_is_provided():
    """
    check that _get_storage_class raises exception if storage class is provided
    that is not a subclass of AbstractTelemetryStorage
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    with pytest.raises(exceptions.StorageClassOfIncorrectType):
        telemetry_inst._get_storage_class(str)


def test_get_storage_class_returns_instance_of_storage_class():
    """
    check that _get_storage_class method returns an instance of
    Storage class provided
    """
    # pylint: disable=too-few-public-methods
    class TestStorage(AbstractTelemetryStorage):
        """test class"""

        def store_telemetry(self, telemetry):
            """test method"""
            return None

    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    assert isinstance(telemetry_inst._get_storage_class(TestStorage), TestStorage)


def test_storage_class_close_method_closes_db():
    """
    check that _get_storage_class method returns an instance of
    Storage class provided
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    storage_instance = telemetry_inst._get_storage_class(None)
    storage_instance.close_db()
    assert storage_instance.db_in_memory is None
    assert storage_instance.db_cursor is None


def test_new_telemetry_has_default_traffic_light_color():
    """New telemetry instance should have default trafic light color."""
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    assert telemetry_inst.traffic_light == DEFAULT_TRAFIC_LIGHT_COLOR


def test_set_telemetry_traffic_light_to_orange():
    """New telemetry instance traffic light property can be set to orange."""
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.set_orange_traffic_light()
    assert telemetry_inst.traffic_light == TRAFIC_LIGHT_COLOR_ORANGE


def test_set_telemetry_traffic_light_to_red():
    """New telemetry instance traffic light property can be set to red."""
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.set_red_traffic_light()
    assert telemetry_inst.traffic_light == TRAFIC_LIGHT_COLOR_RED



"""Telemety Decorators to be used whith class methods that need telemetry

decorators:
    - add_telemetry
    - add_mongo_telemetry
"""
from .main import Telemetry
from .storage import TelemetryMongoStorage


def add_telemetry(telemetry_params: dict) -> object:
    """
    Decorator method to add a telemetry to the class from which the
    decorator was called. When using this decorator the mongo storage
    class will used.

    Args:
        telemetry_params (dict:
            - source_name (str): free format process name
            - process_type (ProcessType): process type definition
            - telemetry_rules (dict): telemetry rules definition dict
            - storage_class (AbstractTelemetryStorage, optional):
                Storage class to be used to store telemetry instances. Defaults
                to TelemetryInMemoryStorage which stores only in memory.
    """

    def wrapper(method):
        def wrapped_method(self, *args, **kwargs):
            """
            Wrapper for method where result log should be added
            """
            if (not hasattr(self, "_telemetry")) or (not self._telemetry):
                self._telemetry = Telemetry(**telemetry_params)
                result = method(self, *args, **kwargs)
                self._telemetry.save_and_close()
                self._telemetry = None
            else:
                result = method(self, *args, **kwargs)

            return result

        return wrapped_method

    return wrapper


def add_mongo_telemetry(telemetry_params: dict) -> object:
    """
    Decorator method to add a telemetry to the class from which the
    decorator was called. When using this decorator the mongo storage
    class will used.

    Args:
        telemetry_params (dict:
            - source_name (str): free format process name
            - process_type (ProcessType): process type definition
            - telemetry_rules (dict): telemetry rules definition dict
    """

    def wrapper(method):
        def wrapped_method(self, *args, **kwargs):
            """
            Wrapper for method where result log should be added
            """
            if (not hasattr(self, "_telemetry")) or (not self._telemetry):
                self._telemetry = Telemetry(**telemetry_params)
                self._telemetry.storage_class = TelemetryMongoStorage
                result = method(self, *args, **kwargs)
                self._telemetry.save_and_close()
                self._telemetry = None
            else:
                result = method(self, *args, **kwargs)
            return result

        return wrapped_method

    return wrapper



"""
module to test entries_have_key instruction class
"""
import pytest

from pipeline_telemetry.settings.exceptions import FieldNameMandatory, \
    MustHaveKeyMandatory
from pipeline_telemetry.validators.entries_have_key import EntriesHaveKey

# pylint: disable=protected-access


def test_has_key_class_exists():
    """ check that HasKey class exists """
    assert EntriesHaveKey


def test_has_key_class_is_singelton():
    """ check that HasKey is a singleton class """
    instance_one = EntriesHaveKey()
    instance_two = EntriesHaveKey()
    assert instance_one is instance_two


def test_validate_rule_method_raises_exception_without_field_name():
    """
    check that _validate_rule method raises exception when field name
    not provided in rule
    """
    rule_content = {'not_field_name': 'test', 'must_have_key': 'test'}
    with pytest.raises(FieldNameMandatory) as exception:
        EntriesHaveKey._validate_rule(rule_content)

    assert 'For `entries_have_key` instruction' in str(exception)


def test_validate_rule_method_raises_exception_without_must_have_key():
    """
    check that _validate_rule method raises exception when the
    must_have_key is not provided in rule
    """
    rule_content = {'field_name': 'test', 'not_must_have_key': 'test'}
    with pytest.raises(MustHaveKeyMandatory) as exception:
        EntriesHaveKey._validate_rule(rule_content)

    assert 'For `entries_have_key` instruction' in str(exception)


def test_validate_rule_method_returns_none():
    """
    check that _validate_rule method returns none rule is valid
    """
    valid_rule_content = {'field_name': 'test', 'must_have_key': 'test'}
    assert EntriesHaveKey._validate_rule(valid_rule_content) is None


def test_get_field_name_method():
    """
    check that _get_field_name method returns value of
    field_name that is in scope of validation
    """
    rule_content = {'field_name': 'test'}
    assert EntriesHaveKey._get_field_name(rule_content) == 'test'


def test_get_must_have_key_method():
    """
    check that _get_must_have_key method returns value of
    must_have_key that is in scope of validation
    """
    rule_content = {'must_have_key': 'test'}
    assert EntriesHaveKey._get_must_have_key(rule_content) == 'test'


def test_validate_field_exists():
    """
    test that _validate_field_exists method returns empty list when
    entries field exists
    """
    dict_to_validate = {'items': [{'key': 1}, {'key': 2}]}
    rule_content = {'field_name': 'items', 'must_have_key': 'key'}

    assert EntriesHaveKey._validate_field_exists(
        dict_to_validate, rule_content) == []


def test_validate_field_exists_returns_error():
    """
    test that _validate_field_exists method returns empty list when
    entries field exists
    """
    dict_to_validate = {'not_items': [{'key': 1}, {'key': 2}]}
    rule_content = {'field_name': 'items', 'must_have_key': 'key'}
    errors = EntriesHaveKey._validate_field_exists(
        dict_to_validate, rule_content)

    assert len(errors) == 1
    assert 'Entries field does not exist' in errors[0].description


def test_validate_entries_have_key():
    """
    test that _validate_entries_have_key method returns empty list when all
    entries have the right key
    """
    dict_to_validate = {'items': [{'key': 1}, {'key': 2}]}
    rule_content = {'field_name': 'items', 'must_have_key': 'key'}

    assert EntriesHaveKey._validate_entries_have_key(
        dict_to_validate, rule_content) == []


def test_validate_entries_have_nested_key():
    """
    test that _validate_entries_have_key method returns empty list when all
    entries have the right key
    """
    dict_to_validate = {'items': [{'key': {'nested_key': 2}}]}
    rule_content = {'field_name': 'items', 'must_have_key': 'key.nested_key'}

    assert EntriesHaveKey._validate_entries_have_key(
        dict_to_validate, rule_content) == []


def test_validate_entries_have_key_with_missing_key():
    """
    test that _validate_entries_have_key method returns list with has_key
    error when a key is missing
    """
    dict_to_validate = {'items': [{'not_key': 1}, {'key': 2}]}
    rule_content = {'field_name': 'items', 'must_have_key': 'key'}

    errors = EntriesHaveKey._validate_entries_have_key(
        dict_to_validate, rule_content)

    assert len(errors) == 1
    assert 'Key missing in entry' in errors[0].description


def test_validate_entries_have_key_with_non_dict_items():
    """
    test that _validate_entries_have_key method returns list with
    entry_is_not_a_dict error when entry is not a dicy
    """
    dict_to_validate = {'items': ['not_a_dict', {'key': 2}]}
    rule_content = {'field_name': 'items', 'must_have_key': 'key'}

    errors = EntriesHaveKey._validate_entries_have_key(
        dict_to_validate, rule_content)

    assert len(errors) == 1
    assert 'Entry that needs to have a key is not a dict' in \
        errors[0].description


def test_validate_entries_have_keys_with_missing_key():
    """
    test that _validate_entries_have_key method returns list with has_key
    errors when a all keys are missing
    """
    dict_to_validate = {'items': [{'not_key': 1}, {'also_not_key': 2}]}
    rule_content = {'field_name': 'items', 'must_have_key': 'key'}

    errors = EntriesHaveKey._validate_entries_have_key(
        dict_to_validate, rule_content)

    assert len(errors) == 2


def test_validate_calls_validate_field_exists(mocker):
    """
    test that _validate method calls the _validate_field_exists method
    """
    dict_to_validate = {'not_items': [{'not_key': 1}, {'also_not_key': 2}]}
    rule_content = {'field_name': 'items', 'must_have_key': 'key'}

    _validate_field_exists_spy = mocker.spy(
        EntriesHaveKey, '_validate_field_exists')

    EntriesHaveKey._validate(dict_to_validate, rule_content)

    assert _validate_field_exists_spy.called


def test_validate_calls_validate_type_entries_field(mocker):
    """
    test that _validate method calls the _validate_type_entries_field method
    """
    dict_to_validate = {'items': 'not a list or a dict'}
    rule_content = {'field_name': 'items', 'must_have_key': 'key'}

    _validate_type_entries_field_spy = mocker.spy(
        EntriesHaveKey, '_validate_type_entries_field')

    EntriesHaveKey._validate(dict_to_validate, rule_content)

    assert _validate_type_entries_field_spy.called


def test_validate_calls_validate_entries_have_key(mocker):
    """
    test that _validate method calls the _validate_type_entries_field method
    """
    dict_to_validate = {'items': [{'a': 1}]}
    rule_content = {'field_name': 'items', 'must_have_key': 'key'}

    _validate_entries_have_key_spy = mocker.spy(
        EntriesHaveKey, '_validate_type_entries_field')

    EntriesHaveKey._validate(dict_to_validate, rule_content)

    assert _validate_entries_have_key_spy.called


def test_validate_calls_validate_rule(mocker):
    """
    test that public validate method calls the _validate_rule method
    """
    dict_to_validate = {'items': [{'a': 1}]}
    rule_content = {'field_name': 'items', 'must_have_key': 'key'}

    _validate_rule_spy = mocker.spy(EntriesHaveKey, '_validate_rule')
    EntriesHaveKey.validate(dict_to_validate, rule_content)

    assert _validate_rule_spy.called


def test_validate_calls_validate(mocker):
    """
    test that public validate method calls the _validate method
    """
    dict_to_validate = {'items': [{'a': 1}]}
    rule_content = {'field_name': 'items', 'must_have_key': 'key'}

    _validate_spy = mocker.spy(EntriesHaveKey, '_validate')
    EntriesHaveKey.validate(dict_to_validate, rule_content)

    assert _validate_spy.called



"""[summary]
"""
import json
import sqlite3

from .generic import AbstractTelemetryStorage


class TelemetryInMemoryStorage(AbstractTelemetryStorage):
    """
    Class to provice telemetry in memory storage for use when unit testing
    """

    db_in_memory = None
    db_cursor = None

    def __init__(self):
        if not self.db_in_memory:
            self.initialize_db()

    @classmethod
    def initialize_db(cls):
        """
        class method to initialize the in memory db
        """
        if not cls.db_in_memory:
            cls.db_in_memory = sqlite3.connect(":memory:")
            cls.db_cursor = cls.db_in_memory.cursor()
            cls._define_db_table(cls.db_cursor)

    @classmethod
    def close_db(cls):
        """close cursor and connections"""
        cls.db_cursor.close()
        cls.db_in_memory.close()
        cls.db_cursor = None
        cls.db_in_memory = None

    @staticmethod
    def _define_db_table(cursor):
        """define telemetry table"""
        cursor.executescript(
            """
            DROP TABLE IF EXISTS telemetry;
            CREATE TABLE telemetry (category varchar(60),
            sub_category varchar(60), source_name varchar(40),
            process_type varchar(40), start_date_time timestamp,
            run_time varchar(20), telemetry_data json,
            traffic_light varchar(10))"""
        )

    def store_telemetry(self, telemetry: dict) -> None:
        """public method to persist telemetry object"""
        telemetry_copy = telemetry.copy()
        category = telemetry_copy.pop("category", None)
        sub_category = telemetry_copy.pop("sub_category", None)
        source_name = telemetry_copy.pop("source_name", None)
        process_type = telemetry_copy.pop("process_type", None)
        start_date_time = telemetry_copy.pop("start_date_time", None)
        run_time_in_seconds = telemetry_copy.pop("run_time_in_seconds", None)
        traffic_light = telemetry_copy.pop("traffic_light", None)
        json_object = json.dumps(telemetry_copy)

        self.db_cursor.execute(
            "insert into telemetry values (?, ?, ?, ?, ?, ?, ?, ?)",
            [
                category,
                sub_category,
                source_name,
                process_type,
                start_date_time,
                run_time_in_seconds,
                json_object,
                traffic_light
            ],
        )



"""Module to define storage class for mongoDB
"""
import os


def get_mongo_db_port():
    """Returns mongo DB port as int or None."""
    mongo_db_port = os.getenv("MONGO_DB_PORT")
    if mongo_db_port:
        return int(mongo_db_port)

    return None


DEFAULT_MONGO_DB_HOST = '127.0.0.1'
DEFAULT_MONGO_DB_PORT = 27017

MONGO_DB_NAME = os.getenv("MONGO_DB_NAME") or 'telemetry'
MONGO_DB_PASSWORD = os.getenv("MONGO_DB_PASSWORD")
MONGO_DB_USERNAME = os.getenv("MONGO_DB_USERNAME")
MONGO_DB_HOST = os.getenv("MONGO_DB_HOST")
MONGO_DB_PORT = get_mongo_db_port()

# if both host and port are not defined the use default mongo host and port
if not (MONGO_DB_HOST or MONGO_DB_PORT):
    MONGO_DB_HOST = DEFAULT_MONGO_DB_HOST
    MONGO_DB_PORT = DEFAULT_MONGO_DB_PORT

MONGO_ACCESS_PARAMS = {
    'db': MONGO_DB_NAME,
    'password': MONGO_DB_PASSWORD,
    'username': MONGO_DB_USERNAME,
    'port': MONGO_DB_PORT,
    'host': MONGO_DB_HOST
}



"""Module to define ProcesTypes class."""
from pipeline_telemetry.settings import exceptions
from pipeline_telemetry.settings.data_class import ProcessType
from pipeline_telemetry.settings.settings import BaseEnumerator


class ProcessTypes():
    """Singleton Class for registering process_types"""

    _process_types = []

    def __new__(cls):
        return cls

    @classmethod
    def register_process_type(
            cls, process_type_key: str, process_type: ProcessType) -> None:
        """Class method to register a single process type."""
        if not isinstance(process_type, ProcessType):
            raise exceptions.ProcessTypeMustBeOfClassProcessType
        setattr(cls, process_type_key, process_type)
        cls._process_types.append(process_type)

    @classmethod
    def register_process_types(cls, process_types: BaseEnumerator) -> None:
        """Class method to register new errors from enumerator."""
        try:
            if not issubclass(process_types, BaseEnumerator):
                raise exceptions.ProcessTypesMustBeOfClassBaseEnumertor
        except TypeError:
            raise exceptions.ProcessTypesMustBeOfClassBaseEnumertor

        for process_type_key in process_types.keys():
            process_type = process_types[process_type_key].value
            cls.register_process_type(
                process_type_key=process_type_key, process_type=process_type)

    @classmethod
    def is_registered(cls, process_type: ProcessType) -> bool:
        """Method checks of a process_type is registered.

        Args:
            process_type (ProcessType): ProcessType instance to be checked

        Returns:
            Bool: True of process_type is registered else False
        """
        return process_type in cls._process_types



"""Module to define test data for the integration tests
"""
from dataclasses import dataclass

from pipeline_telemetry.main import ERRORS_KEY


@dataclass(frozen=True)
class IntegrationTestData:
    """Immutable dataclass to define integration test data."""

    data: dict
    validation_rules: dict
    telemetry_result: dict

    @property
    def result_field(self):
        return self.telemetry_result.get("field")

    @property
    def expected_result(self):
        return self.telemetry_result.get("result")


def validate_result_from_telemetry(
    telemetry: dict, test_data: IntegrationTestData
) -> bool:
    """
    Validates if the telemetry data is according to expected results as defined
    in IntegrationTestData object test_data
    """
    result_field = test_data.result_field
    expected_result = test_data.expected_result
    return telemetry.get(result_field) == expected_result


BASIC_RETRIEVE_DATA_TEST = IntegrationTestData(
    data={
        "items": [
            {"id": 1, "field1": "test"},
            {"id": 2, "field1": "test"},
            {"id": 3, "field1": "test"},
            {"id": 4, "field1": "test"},
        ]
    },
    validation_rules={
        "RETRIEVE_RAW_DATA": {
            "has_key": {"field_name": "items"},
            "validate_entries": {"field_name": "items", "expected_count": 4},
        }
    },
    telemetry_result={
        "field": "RETRIEVE_RAW_DATA",
        "result": {"base_counter": 1, "fail_counter": 0, ERRORS_KEY: {}},
    },
)

MISSING_KEY_RETRIEVE_DATA_TEST = IntegrationTestData(
    data={"no items key": []},
    validation_rules={"RETRIEVE_RAW_DATA": {"has_key": {"field_name": "items"}}},
    telemetry_result={
        "field": "RETRIEVE_RAW_DATA",
        "result": {
            "base_counter": 1,
            "fail_counter": 0,
            ERRORS_KEY: {"HAS_KEY_ERR_0001@KEY_<items>": 1},
        },
    },
)


MUST_HAVE_KEY_IN_ITEMS_TEST = IntegrationTestData(
    data={"items": [{"key": 1}, {"no_key": 2}]},
    validation_rules={
        "RETRIEVE_RAW_DATA": {
            "entries_have_key": {"field_name": "items", "must_have_key": "key"}
        }
    },
    telemetry_result={
        "field": "RETRIEVE_RAW_DATA",
        "result": {
            "base_counter": 1,
            "fail_counter": 0,
            ERRORS_KEY: {"ENTRIES_HAVE_KEY_ERR_003@KEY_<items_key>": 1},
        },
    },
)

NO_ITEMS_HAVE_MUST_HAVE_KEY_TEST = IntegrationTestData(
    data={"items": [{"no_key": 1}, {"no_key": 2}]},
    validation_rules={
        "RETRIEVE_RAW_DATA": {
            "entries_have_key": {"field_name": "items", "must_have_key": "key"}
        }
    },
    telemetry_result={
        "field": "RETRIEVE_RAW_DATA",
        "result": {
            "base_counter": 1,
            "fail_counter": 0,
            ERRORS_KEY: {"ENTRIES_HAVE_KEY_ERR_003@KEY_<items_key>": 2},
        },
    },
)

ITEMS_FIELD_HAS_DICT_TEST = IntegrationTestData(
    data={"items": {"key": 1, "no_key": 2}},
    validation_rules={
        "RETRIEVE_RAW_DATA": {
            "entries_have_key": {"field_name": "items", "must_have_key": "key"}
        }
    },
    telemetry_result={
        "field": "RETRIEVE_RAW_DATA",
        "result": {
            "base_counter": 1,
            "fail_counter": 0,
            ERRORS_KEY: {"ENTRIES_HAVE_KEY_ERR_002@KEY_<items>": 1},
        },
    },
)



"""
module to test custom exceptions in pipeline telemetry module
"""
import pytest
from test_data import TEST_PROCESS_TYPE

from pipeline_telemetry.settings import exceptions


def raise_exception(exception, value):
    """ test helper method to raise exception """
    if value:
        raise exception(value)
    raise exception


def test_field_name_mandatory_exception():
    """ test FieldNameMandatory exception """
    with pytest.raises(exceptions.FieldNameMandatory) as exception:
        raise_exception(exceptions.FieldNameMandatory, 'test_field_name')

    assert 'For `test_field_name` instruction' in str(exception)


def test_unknown_instruction_exception():
    """ test UnknownInstruction exception """
    with pytest.raises(exceptions.UnknownInstruction) as exception:
        raise_exception(exceptions.UnknownInstruction, 'test_instruction')

    assert 'test_instruction not registered' in str(exception)


def test_instruction_registered_twice_exception():
    """ test InstructionRegisteredTwice exception """
    class TestInstructrion():
        """ test class """
        instruction = 'test_instruction'

    with pytest.raises(exceptions.InstructionRegisteredTwice) as exception:
        raise_exception(
            exceptions.InstructionRegisteredTwice, TestInstructrion)

    assert 'test_instruction from class TestInstructrion.' in str(exception)


def test_rule_can_only_have_one_instruction_exception():
    """ test RuleCanHaveOnlyOneInstruction exception """
    with pytest.raises(exceptions.RuleCanHaveOnlyOneInstruction) as exception:
        raise_exception(
            exceptions.RuleCanHaveOnlyOneInstruction, {'key1': 1, 'key2': 2})

    assert 'Rule contains multiple keys' in str(exception)
    assert 'key1, key2' in str(exception)


def test_base_count_for_sub_process_not_added_exception():
    """ test BaseCountForSubProcessNotAdded exception """
    with pytest.raises(exceptions.BaseCountForSubProcessNotAdded) as exception:
        raise_exception(
            exceptions.BaseCountForSubProcessNotAdded, 'sub_process')

    assert 'Sub process sub_process has not yet been initialized' in \
        str(exception)


def test_telemetry_object_already_closed_exception():
    """ test TelemetryObjectAlreadyClosed exception """
    with pytest.raises(exceptions.TelemetryObjectAlreadyClosed) as exception:
        raise_exception(exceptions.TelemetryObjectAlreadyClosed, None)

    assert 'Telemetry object is already closed' in str(exception)


def test_storage_class_of_incorrect_type_exception():
    """ test StorageClassOfIncorrectType exception """
    with pytest.raises(exceptions.StorageClassOfIncorrectType) as exception:
        raise_exception(
            exceptions.StorageClassOfIncorrectType, 'Incorrect_class')

    assert 'StorageClass `Incorrect_class` not a child class' in str(exception)


def test_expected_count_must_be_positive_int_exception():
    """ test ExpectedCountMustBePositiveInt exception """
    with pytest.raises(exceptions.ExpectedCountMustBePositiveInt) as exception:
        raise exceptions.ExpectedCountMustBePositiveInt

    assert 'Ruleset does not contain `expected_count` or' in str(exception)


def test_invalid_sub_process():
    """ test InvalidSubProcess exception """
    with pytest.raises(exceptions.InvalidSubProcess) as exception:
        raise exceptions.InvalidSubProcess('test_process', TEST_PROCESS_TYPE)

    assert 'Sub Process `test_process` does not exist' in str(exception)


def test_process_type_not_registered():
    """ test ProcessTypeNotRegistered exception """
    with pytest.raises(exceptions.ProcessTypeNotRegistered) as exception:
        raise exceptions.ProcessTypeNotRegistered(TEST_PROCESS_TYPE)

    assert 'provided process_type test_process_type not registered' in \
        str(exception)


def test_process_types_not_of_right_class():
    """ test ProcessTypesMustBeOfClassBaseEnumertor exception """
    with pytest.raises(
            exceptions.ProcessTypesMustBeOfClassBaseEnumertor) as exception:
        raise exceptions.ProcessTypesMustBeOfClassBaseEnumertor

    assert 'provided process_types enumerator not of class BaseEnumrator' in \
        str(exception)


def test_sub_process_not_defined_in_process_type():
    """ test InvalidSubProcessForProcessType exception """
    with pytest.raises(
            exceptions.InvalidSubProcessForProcessType) as exception:
        raise exceptions.InvalidSubProcessForProcessType

    assert 'Provided sub process is not defined in ProcessType.' in \
        str(exception)



"""
Module to test telemetry counter login for pipeline telemetry module.
"""
import pytest
from test_data import DEFAULT_TELEMETRY_PARAMS, TEST_ERROR_TELEMETRY_COUNTER, \
    TEST_INV_TELEMETRY_COUNTER, TEST_TELEMETRY_COUNTER, \
    TEST_TELEMETRY_COUNTER_INC_2

from pipeline_telemetry import ProcessTypes
from pipeline_telemetry.helper import is_telemetry_counter
from pipeline_telemetry.main import ERRORS_KEY, Telemetry
from pipeline_telemetry.settings import exceptions
from pipeline_telemetry.settings.data_class import TelemetryCounter


@pytest.fixture
def telemetry_inst():
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry_inst.increase_sub_process_base_count("RETRIEVE_RAW_DATA")
    return telemetry_inst


def test_telemetry_counter_exists():
    """check that TelemetryCounter class exists"""
    assert TelemetryCounter


def test_add_telemetry_counter_to_telemetry(telemetry_inst):
    """
    Check that adding a telemetry counter to a Telemetry instance
    creates a counter
    """
    telemetry_inst.add_telemetry_counter(TEST_TELEMETRY_COUNTER)
    assert telemetry_inst.get("RETRIEVE_RAW_DATA")["test_counter"] == 1


def test_add_error_telemetry_counter_to_telemetry(telemetry_inst):
    """
    Check that adding a telemetry counter to a Telemetry instance
    creates a counter
    """
    telemetry_inst.add_telemetry_counter(TEST_ERROR_TELEMETRY_COUNTER)
    assert telemetry_inst.get("RETRIEVE_RAW_DATA")[ERRORS_KEY]["HAS_KEY_ERR_0001"] == 1


def test_add_error_telemetry_counter_to_telemetry_raises_exception():
    """
    Check that adding a telemetry counter to a Telemetry instance
    raises exception when basecounter has not been added
    """
    telemetry_inst = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    with pytest.raises(exceptions.BaseCountForSubProcessNotAdded):
        telemetry_inst.add_telemetry_counter(TEST_ERROR_TELEMETRY_COUNTER)


def test_add_telemetry_counter_with_icrement_2_to_telemetry(telemetry_inst):
    """
    Check that adding a telemetry counter with increment = 2 to a Telemetry
    raises the counter with value of 2
    """
    telemetry_inst.add_telemetry_counter(TEST_TELEMETRY_COUNTER)
    assert telemetry_inst.get("RETRIEVE_RAW_DATA")["test_counter"] == 1
    telemetry_inst.add_telemetry_counter(TEST_TELEMETRY_COUNTER_INC_2)
    assert telemetry_inst.get("RETRIEVE_RAW_DATA")["test_counter"] == 3


def test_add_telemetry_counter_with_increment_arg(telemetry_inst):
    """
    Check invoking add_telemetry_counter method using increment argument adds
    the value of increment to the counter
    """
    telemetry_inst.add_telemetry_counter(TEST_TELEMETRY_COUNTER, increment=10)
    assert telemetry_inst.get("RETRIEVE_RAW_DATA")["test_counter"] == 10


def test_add_telemetry_counter_raises_exception(telemetry_inst):
    """
    Check invoking add_telemetry_counter method using increment argument adds
    the value of increment to the counter
    """
    with pytest.raises(exceptions.InvalidSubProcessForProcessType):
        telemetry_inst.add_telemetry_counter(TEST_INV_TELEMETRY_COUNTER)


def test_is_telemetry_counter_returns_false():
    """
    Test is_telemetry_counter method returns False if counter is not an instance
    off TelemetryCounter or subclass of TelemetryCounter.
    """
    assert not is_telemetry_counter("1")


def test_is_telemetry_counter_returns_true():
    """
    Test is_telemetry_counter method returns True if counter is an instance of
    TelemetryCounter.
    """
    assert is_telemetry_counter(TEST_TELEMETRY_COUNTER)


def test_is_error_returns_true_on_error_code_sub_class():
    """
    Test is_error method returns True if error is an instance of ErrorCode.
    """

    class TelemetryCounterSubClass(TelemetryCounter):
        pass

    counter = TelemetryCounterSubClass(
        process_type=TEST_TELEMETRY_COUNTER.process_type,
        sub_process=TEST_TELEMETRY_COUNTER.sub_process,
        counter_name=TEST_TELEMETRY_COUNTER.counter_name,
    )

    assert is_telemetry_counter(counter)


def test_telemetry_counter_validate_sub_process():
    """
    Test validate_sub_process method returns None if sub process is in scope of
    process_type.
    """
    assert TEST_TELEMETRY_COUNTER.validate_sub_process() is None


def test_validate_sub_process_raises_exception():
    """
    Test validate_sub_process method raises exception if sub process is not
    in scope of process_type.
    """
    counter = TelemetryCounter(
        process_type=TEST_TELEMETRY_COUNTER.process_type,
        sub_process="sub process not in scope of process_type",
        counter_name=TEST_TELEMETRY_COUNTER.counter_name,
    )

    with pytest.raises(exceptions.InvalidSubProcessForProcessType):
        counter.validate_sub_process()


def test_validate_sub_process_with_multiple_process_types():
    """
    Test validate_sub_process method returns None if sub process is in scope of
    one of the given process_types.
    """
    counter = TelemetryCounter(
        process_types=[ProcessTypes.CREATE_DATA_FROM_URL, ProcessTypes.UPLOAD_DATA],
        sub_process="DATA_UPLOAD",
        counter_name=TEST_TELEMETRY_COUNTER.counter_name,
    )
    assert counter.validate_sub_process() is None


def test_all_process_types_with_multiple_entries():
    """
    Test validate_sub_process method returns None if sub process is in scope of
    one of the given process_types.
    """
    counter = TelemetryCounter(
        process_types=[ProcessTypes.CREATE_DATA_FROM_URL, ProcessTypes.UPLOAD_DATA],
        sub_process="DATA_UPLOAD",
        counter_name=TEST_TELEMETRY_COUNTER.counter_name,
    )
    assert None not in counter.all_process_types
    assert ProcessTypes.CREATE_DATA_FROM_URL in counter.all_process_types
    assert ProcessTypes.UPLOAD_DATA in counter.all_process_types


def test_all_process_types_with_single_entry():
    """
    Test validate_sub_process method returns None if sub process is in scope of
    one of the given process_types.
    """
    counter = TelemetryCounter(
        process_type=ProcessTypes.CREATE_DATA_FROM_URL,
        sub_process="DATA_UPLOAD",
        counter_name=TEST_TELEMETRY_COUNTER.counter_name,
    )
    assert None not in counter.all_process_types
    assert ProcessTypes.CREATE_DATA_FROM_URL in counter.all_process_types



"""Module to define EntriesHaveKey class validator
"""
import jmespath
from errors import ErrorCode, ListErrors

from ..settings import exceptions
from ..validators.dict_validator import DictValidator


class EntriesHaveKey():
    """
    class to define HasFieldName instruction. Validator checks if all
    entries in a field contain a specific key.

    Validator can be invoked with command `entries_have_key`
    rule_content needs to contain:
        - field_name (str): field that holds the entries in a dict or a list
        - must_have_key (str): field that hold the key that is to be present
                               in all entries

    public method
        - validate
    """
    instruction = 'entries_have_key'
    fieldname = 'field_name'
    must_have_key = 'must_have_key'

    def __new__(cls):
        """ make this a singleton class """
        return cls

    @classmethod
    def validate(
            cls, dict_to_validate: dict, rule_content: dict) -> list[ErrorCode]:
        """
        Public method run the validation
        prior to validation the rule will be validated.

        Args:
            dict_to_validate (dict): Input dict that needs validation
            rule_content (dict): actual rule

        Returns:
            list[ErrorCode]: List with errors that were found. Empty list if no
                             errors were found

        Raise:
            FieldNameMandatory: when field name not defined in rule
            MustHaveKeyMandatory: when must have key is not defined in rule
        """
        cls._validate_rule(rule_content)
        return cls._validate(dict_to_validate, rule_content)

    @classmethod
    def _validate(
            cls, dict_to_validate: dict,
            rule_content: dict) -> list[ErrorCode]:
        """
        method to do the actual validation

        returns:
            - error in case
        """
        # if entries does not exist stop validation and return error
        if error := cls._validate_field_exists(dict_to_validate, rule_content):
            return error

        # if entries field is notlist or dict stop validation and return error
        if error := cls._validate_type_entries_field(
                dict_to_validate, rule_content):
            return error

        # return the errors per entry
        return cls._validate_entries_have_key(dict_to_validate, rule_content)

    @classmethod
    def _validate_entries_have_key(
            cls, dict_to_validate: dict,
            rule_content: dict) -> list[ErrorCode]:
        """
        method to check if the values in the entry have a specific key

        returns:
            - list{ErrorCode]: Empty list in case all entries have the key
        """
        errors = []
        fieldname = cls._get_field_name(rule_content)
        must_have_key = cls._get_must_have_key(rule_content)

        # prepare error messages
        entry_is_not_a_dict_error = cls._validation_error(
            ListErrors.ENTRY_IS_NOT_A_DICT, fieldname)
        missing_key_in_entry_error = cls._validation_error(
            ListErrors.KEY_NOT_FOUND_IN_ENTRY, fieldname + "_" + must_have_key)

        # for each entry do a has_key validation with the `must_have_key` field
        for entry in jmespath.search(fieldname, dict_to_validate):
            if not isinstance(entry, dict):
                errors.extend(entry_is_not_a_dict_error)
                continue

            if not jmespath.search(must_have_key, entry):
                errors.extend(missing_key_in_entry_error)

        return errors

    @classmethod
    def _validate_field_exists(
            cls, dict_to_validate: dict,
            rule_content: dict) -> list[ErrorCode]:
        """
        method to check if entries field exists in dict_to_validate

        returns:
            - list{ErrorCode]: Empty list in case the field does exist
        """
        fieldname = cls._get_field_name(rule_content)
        if not jmespath.search(fieldname, dict_to_validate):
            return cls._validation_error(
                ListErrors.ENTRIES_FIELD_NOT_FOUND, fieldname)

        return []

    @classmethod
    def _validate_type_entries_field(
            cls, dict_to_validate: dict,
            rule_content: dict) -> list[ErrorCode]:
        """
        method to check if entries field in dict_to_validate is a list

        returns:
            - list[ErrorCode]: Empty list in case the field holds a list
        """
        fieldname = cls._get_field_name(rule_content)
        entries = jmespath.search(fieldname, dict_to_validate)
        if not isinstance(entries, (list)):
            return cls._validation_error(
                ListErrors.ENTRIES_FIELD_OF_WRONG_TYPE, fieldname)

        return []

    @classmethod
    def _validate_rule(cls, rule_content: dict) -> None:
        """Method to raise exception when rule_content in invalid."""
        fieldname = cls._get_field_name(rule_content)
        if not isinstance(fieldname, str):
            raise exceptions.FieldNameMandatory(cls.instruction)

        must_have_key = cls._get_must_have_key(rule_content)
        if not isinstance(must_have_key, str):
            raise exceptions.MustHaveKeyMandatory(cls.instruction)

    @classmethod
    def _get_must_have_key(cls, rule_content: dict) -> str:
        """
        retrieves the must_have_key from the rule content
        """
        return rule_content.get(cls.must_have_key, None)

    @classmethod
    def _get_field_name(cls, rule_content: dict) -> str:
        """
        retrieves the field name in scope from the rule content
        """
        return rule_content.get(cls.fieldname, None)

    @staticmethod
    def _validation_error(
            error_code: ErrorCode, fieldname: str) -> list[ErrorCode]:
        """ returns error code object in list with fieldname as error data """
        key_specific_error = error_code.code + '@KEY_<' + fieldname + '>'
        return [ErrorCode(
            code=key_specific_error, description=error_code.description)]


DictValidator.register_instruction(EntriesHaveKey)



"""
Module to define customer excpetions for data_validator module

exceptions
- FieldNameMandatory
- UnknownInstruction
- InstructionRegisteredTwice
- RuleCanHaveOnlyOneInstruction
- BaseCountForSubProcessNotAdded
- TelemetryObjectAlreadyClosed
- StorageClassOfIncorrectType
- ProcessTypeMustBeOfClassProcessType
- ProcessTypeNotRegistered
"""


class FieldNameMandatory(Exception):
    """custom exception for Telemetry Module"""

    def __init__(self, instruction):
        message = f"For `{instruction}` instruction"
        super().__init__(message)


class MustHaveKeyMandatory(Exception):
    """custom exception for Telemetry Module"""

    def __init__(self, instruction):
        message = f"For `{instruction}` instruction"
        super().__init__(message)


class ExpectedCountMustBePositiveInt(Exception):
    """custom exception for Telemetry Module"""

    def __init__(self):
        message = "".join(
            [
                "Ruleset does not contain `expected_count` or ",
                "expected count is negative",
            ]
        )
        super().__init__(message)


class UnknownInstruction(Exception):
    """custom exception for Telemetry Module"""

    def __init__(self, instruction):
        super().__init__(f"{instruction} not registered")


class InstructionRegisteredTwice(Exception):
    """custom exception for Telemetry Module"""

    def __init__(self, instruction_class):
        message = "".join(
            [
                f"{instruction_class.instruction} from class ",
                f"{instruction_class.__name__}.",
            ]
        )
        super().__init__(message)


class RuleCanHaveOnlyOneInstruction(Exception):
    """custom exception for Telemetry Module"""

    def __init__(self, rule):
        message = "".join(
            ["Rule contains multiple keys / instruction : ", ", ".join(rule.keys())]
        )
        super().__init__(message)


class InvalidSubProcess(Exception):
    """custom exception for Telemetry Module"""

    def __init__(self, sub_process: str, process_type):
        message = "".join(
            [
                f"Sub Process `{sub_process}` does not exist for ",
                f"process type `{process_type.name}`",
            ]
        )
        super().__init__(message)


class BaseCountForSubProcessNotAdded(Exception):
    """custom exception for Telemetry Module"""

    def __init__(self, sub_process):
        message = "".join(
            [
                f"Sub process {sub_process} has not yet been initialized.",
                f" First call increase_sub_process_base_count({sub_process}) ",
                "on telemetry instance",
            ]
        )
        super().__init__(message)


class TelemetryObjectAlreadyClosed(Exception):
    """custom exception for Telemetry Module"""

    def __init__(self):
        message = "Telemetry object is already closed"
        super().__init__(message)


class StorageClassOfIncorrectType(Exception):
    """custom exception for Telemetry Module"""

    def __init__(self, class_name):
        message = "".join(
            [
                f"StorageClass `{class_name}` not a child ",
                "class of AbstractTelemetryStorage",
            ]
        )
        super().__init__(message)


class ProcessTypeMustBeOfClassProcessType(Exception):
    """custom exception for Telemetry Module"""

    def __init__(self):
        message = "provide process_type not of class ProcessType"
        super().__init__(message)


class ProcessTypeNotRegistered(Exception):
    """custom exception for Telemetry Module"""

    def __init__(self, process_type):
        message = f"provided process_type {process_type.name} not registered"
        super().__init__(message)


class ProcessTypesMustBeOfClassBaseEnumertor(Exception):
    """custom exception for Telemetry Module"""

    def __init__(self):
        message = "provided process_types enumerator not of class BaseEnumrator"
        super().__init__(message)


class InvalidSubProcessForProcessType(Exception):
    def __init__(self):
        message = "Provided sub process is not defined in ProcessType."
        super().__init__(message)



"""
module to test the class methods fro Telemetry class of telemetry module

"""
import pytest

from pipeline_telemetry.main import Telemetry
from pipeline_telemetry.settings import exceptions
from pipeline_telemetry.settings.data_class import ProcessType


def test_telemetry_class_method_add_process_type():
    """
    check that Telemetry class method add_process_type adds a process type
    """
    process_type = ProcessType(
        process_type='custom_process_type',
        subtypes=['custom_sub_process_type'])
    Telemetry.add_process_type('TEST_TYPE', process_type)
    assert Telemetry('test', 'test', 'test', process_type)


def test_sub_process_types_also_added():
    """
    check that Telemetry class method add_process_type adds a process type
    including all sub_processes.
    """
    process_type = ProcessType(
        process_type='custom_process_type',
        subtypes=['custom_sub_process_type'])
    Telemetry.add_process_type('TEST_TYPE', process_type)
    telemetry = Telemetry('category', 'sub_category', 'name', process_type)
    assert telemetry._process_type.sub_processes == ['custom_sub_process_type']


def test_add_process_type_raises_exception():
    """
    test class method add_process_type raises an excpetion when an object
    other the a ProcessType is offered
    """
    with pytest.raises(exceptions.ProcessTypeMustBeOfClassProcessType):
        Telemetry.add_process_type('TYPE_NAME', 'not a ProcessType')



"""
This module provides test data for the telemetry tests
"""
from errors import ErrorCode, ListErrors

from pipeline_telemetry.settings.data_class import ProcessType, \
    TelemetryCounter
from pipeline_telemetry.settings.process_type import ProcessTypes
from pipeline_telemetry.settings.settings import \
    DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES


class InstructionTestClass:
    """Instruction class for test purposes."""

    # pylint: disable=too-few-public-methods
    instruction = "test_instruction"
    fieldname = "field_name"

    def __new__(cls):
        """make this a singleton class"""
        return cls


# default params for creating a test Telemetry object
DEFAULT_TELEMETRY_PARAMS = {
    "category": "WEATHER",
    "sub_category": "DAILY_PREDICTIONS",
    "source_name": "load_weather_data",
    "process_type": ProcessTypes.CREATE_DATA_FROM_URL,
}

# custom process type for testing extending the default process types
CUSTOM_PROCESS_TYPE = {"custom_process_type": ["CUSTOM_SUB_TYPE"]}

TEST_ERROR_CODE = ErrorCode(code="TEST_CODE_0001", description="test error")

TEST_TELEMETRY_RULES = {"RETRIEVE_RAW_DATA": {"has_key": {"field_name": "items"}}}

TEST_PROCESS_TYPE = ProcessType(
    process_type="test_process_type", subtypes=DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES
)

TEST_TELEMETRY_COUNTER = TelemetryCounter(
    process_type=TEST_PROCESS_TYPE,
    sub_process="RETRIEVE_RAW_DATA",
    counter_name="test_counter",
)

TEST_TELEMETRY_COUNTER_INC_2 = TelemetryCounter(
    process_type=TEST_PROCESS_TYPE,
    sub_process="RETRIEVE_RAW_DATA",
    counter_name="test_counter",
    increment=2,
)

TEST_INV_TELEMETRY_COUNTER = TelemetryCounter(
    process_type=TEST_PROCESS_TYPE,
    sub_process="invalid_sub_process",
    counter_name="test_counter",
    increment=2,
)

TEST_ERROR_TELEMETRY_COUNTER = TelemetryCounter(
    process_type=TEST_PROCESS_TYPE,
    sub_process="RETRIEVE_RAW_DATA",
    increment=1,
    error=ListErrors.KEY_NOT_FOUND,
)



"""[summary]
"""
from dataclasses import dataclass
from typing import Any, List

from errors import ErrorCode

from pipeline_telemetry.settings import exceptions


@dataclass(frozen=True)
class ProcessType:
    """Immutable dataclass to define a process type and its subtypes."""

    process_type: str
    subtypes: list = list

    @property
    def name(self) -> str:
        """Property method to return process key name."""
        return self.process_type

    @property
    def sub_processes(self) -> List[str]:
        """Property method to return list of sub process types."""
        return self.subtypes


@dataclass(frozen=True)
class TelemetryCounter:
    """Immutable dataclass to define a process type and its subtypes."""

    sub_process: str
    process_types: List[ProcessType] = None
    process_type: ProcessType = None
    counter_name: str = None
    increment: int = 1
    error: ErrorCode = None

    def add_to(self, object_with_telemetry: Any, increment: int = None) -> None:
        """
        Method to add to add self (the TelemetryCounter) to an object telemetry
        instance
        """
        object_with_telemetry._telemetry.add_telemetry_counter(
            telemetry_counter=self, increment=increment
        )

    def validate_sub_process(self) -> None:
        """
        Raises exception if sub_process not define in ProcessType in scope.
        """
        subtypes = []
        for process_type in self.all_process_types:
            subtypes.extend(process_type.subtypes)

        if self.sub_process not in subtypes:
            raise exceptions.InvalidSubProcessForProcessType

    @property
    def all_process_types(self) -> List[ProcessType]:
        """Returns all process_types in scope in a list

        Returns:
            List[ProcessType]: list with process_types
        """
        all_process_types = [self.process_type]
        if self.process_types:
            all_process_types.extend(self.process_types)

        return list(
            filter(lambda process_type: process_type is not None, all_process_types)
        )



"""Module to test processtype logic.
"""
import pytest
from test_data import TEST_PROCESS_TYPE

from pipeline_telemetry.settings import exceptions
from pipeline_telemetry.settings.process_type import ProcessTypes
from pipeline_telemetry.settings.settings import \
    DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES


def test_process_type_class_exists():
    """Test that process types class exists."""
    assert ProcessTypes


def test_process_type_class_is_a_singleton_class():
    """Test that process types class exists."""
    assert ProcessTypes() is ProcessTypes()


def test_process_type_class_can_call_a_process_type():
    """Test that a process type can be called upon the class exists."""
    assert ProcessTypes.CREATE_DATA_FROM_URL


def test_calling_name_property_on_a_process_type_returns_a_str():
    """
    Test that calling name property on a process type returns the
    process_type name.
    """
    assert ProcessTypes.CREATE_DATA_FROM_URL.name == \
        'create_data_from_url'


def test_calling_sub_processes_property_on_a_process_type_returns_a_list():
    """
    Test that calling sub_processes property on a process type returns the
    list of sub_processes.
    """
    assert ProcessTypes.CREATE_DATA_FROM_URL.sub_processes == \
        DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES


def test_registering_process_type_success():
    """
    Test that a processtype can be registered
    """
    assert not ProcessTypes.is_registered(TEST_PROCESS_TYPE)
    ProcessTypes.register_process_type('TEST_PROCESS_TYPE', TEST_PROCESS_TYPE)
    assert ProcessTypes.is_registered(TEST_PROCESS_TYPE)


def test_registering_process_type_raises_exception():
    """
    Test that registering a processtype with a processtype of wrong class
    raises an exception.
    """
    with pytest.raises(exceptions.ProcessTypeMustBeOfClassProcessType):
        ProcessTypes.register_process_type('TEST_PROCESS_TYPE', 'process_type')


def test_registering_process_types_with_class_instance_raises_exception():
    """
    Test that registering a process types enumerator with a class instance    raises an exception.
    """
    with pytest.raises(exceptions.ProcessTypesMustBeOfClassBaseEnumertor):
        ProcessTypes.register_process_types('process_type')


def test_registering_process_types_with_wrong_class_raises_exception():
    """
    Test that registering a process types enumerator with a wrong class
    raises an exception.
    """
    with pytest.raises(exceptions.ProcessTypesMustBeOfClassBaseEnumertor):
        ProcessTypes.register_process_types(str)



"""Module to define HasKey class validator
"""
import jmespath
from errors import ErrorCode, ListErrors

from ..settings import exceptions
from ..validators.dict_validator import DictValidator


class HasKey():
    """
    class to define HasFieldName instruction
    """
    instruction = 'has_key'
    fieldname = 'field_name'

    def __new__(cls):
        """ make this a singleton class """
        return cls

    @classmethod
    def validate(
            cls, dict_to_validate: dict, rule_content: dict) -> list[ErrorCode]:
        """
        Public method run the validation
        prior to validation the rule will be validated.

        Args:
            dict_to_validate (dict): Input dict that needs validation
            rule_content (dict): actual rule

        Returns:
            list[ErrorCode]: List with errors that were found

        Raise:
            FieldNameMandatory: when field name not defined in rule
        """
        cls._validate_rule(rule_content)
        return cls._validate(dict_to_validate, rule_content)

    @classmethod
    def _validate(
            cls, dict_to_validate: dict,
            rule_content: dict) -> list[ErrorCode]:
        """
        method to do the actual validation

        returns:
            - error in case
        """
        fieldname = cls._get_field_name(rule_content)
        if not jmespath.search(fieldname, dict_to_validate):
            return cls._validation_error(
                ListErrors.KEY_NOT_FOUND, fieldname)

        return []

    @classmethod
    def _validate_rule(cls, rule_content):
        fieldname = cls._get_field_name(rule_content)
        if not isinstance(fieldname, str):
            raise exceptions.FieldNameMandatory(cls.instruction)

    @classmethod
    def _get_field_name(cls, rule_content):
        """
        retrieves the field name in scope from the rule content
        """
        return rule_content.get(cls.fieldname, None)

    @staticmethod
    def _validation_error(
            error_code: ErrorCode, fieldname: str) -> list[ErrorCode]:
        """ returns error code object in list with fieldname as error data """
        key_specific_error = error_code.code + '@KEY_<' + fieldname + '>'
        return [ErrorCode(
            code=key_specific_error, description=error_code.description)]


DictValidator.register_instruction(HasKey)



"""Module to define errors
"""
from errors import ErrorCode
from errors.base import FunctionalErrorsBaseClass


class ValidationErrors(FunctionalErrorsBaseClass):
    """
    Class to define enumerator errors for Error module
    """
    KEY_NOT_FOUND = ErrorCode(
        code='HAS_KEY_ERR_0001',
        description='Key missing in provided dict')

    FIELD_NOT_FOUND = ErrorCode(
        code='VALIDATE_ENTRIES_ERR_001',
        description='Field that needs counting can not be found')

    WRONG_TYPE_IN_FIELD = ErrorCode(
        code='VALIDATE_ENTRIES_ERR_002',
        description='Field does not contain dict or list')

    UNEXPECTED_NR_OF_ITEMS = ErrorCode(
        code='VALIDATE_ENTRIES_ERR_003',
        description='Field contains unexpected nr of items')

    ENTRIES_FIELD_NOT_FOUND = ErrorCode(
        code='ENTRIES_HAVE_KEY_ERR_001',
        description='Entries field does not exist')

    ENTRIES_FIELD_OF_WRONG_TYPE = ErrorCode(
        code='ENTRIES_HAVE_KEY_ERR_002',
        description='Entries field does not contain a list')

    KEY_NOT_FOUND_IN_ENTRY = ErrorCode(
        code='ENTRIES_HAVE_KEY_ERR_003',
        description='Key missing in entry')

    ENTRY_IS_NOT_A_DICT = ErrorCode(
        code='ENTRIES_HAVE_KEY_ERR_004',
        description='Entry that needs to have a key is not a dict')



"""[summary]
"""
from enum import Enum

from pipeline_telemetry.settings.data_class import ProcessType

TRAFIC_LIGHT_COLOR_GREEN = 'GREEN'
TRAFIC_LIGHT_COLOR_ORANGE = 'ORANGE'
TRAFIC_LIGHT_COLOR_RED = 'RED'
DEFAULT_TRAFIC_LIGHT_COLOR = TRAFIC_LIGHT_COLOR_GREEN

DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES = [
    "RETRIEVE_RAW_DATA",
    "DATA_CONVERSION",
    "DATA_STORAGE",
]

DEFAULT_UPLOAD_DATA_SUB_PROCESS_TYPES = [
    "DATA_SELECTION",
    "DATA_CONVERSION",
    "DATA_UPLOAD",
]


class BaseEnumerator(Enum):
    """
    Class to define base enumerator with a keys method.
    Only to be used for defining new enumerators.
    """

    @classmethod
    def keys(cls):
        return list(cls.__members__.keys())


class DefaultProcessTypes(BaseEnumerator):
    """
    Class to define the defailt process types with their subtypes
    """

    CREATE_DATA_FROM_URL = ProcessType(
        process_type="create_data_from_url",
        subtypes=DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES,
    )

    CREATE_DATA_FROM_API = ProcessType(
        process_type="create_data_from_api",
        subtypes=DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES,
    )

    CREATE_DATA_FROM_FILE = ProcessType(
        process_type="create_data_from_file",
        subtypes=DEFAULT_CREATE_DATA_SUB_PROCESS_TYPES,
    )

    UPLOAD_DATA = ProcessType(
        process_type="upload_data", subtypes=DEFAULT_UPLOAD_DATA_SUB_PROCESS_TYPES
    )



"""
Module pipeline_telemetry
Providing Telemetry class and functionality to store pipeline telemetry detals
Main usage:

    >>> from pipeline_telemetry.main import Telemetry
    >>> telemetry = Telemetry(
            source_name, process_type, telemetry_rules, storage_class)
    >>> pipeline_telemetry.add(sub_process_type, data, errors)
    >>> pipeline_telemetry.save_and_close()

In this example a telemetry object is created, the result of a sub_process is
then added with either the data resulting from this sub process or the error
that the sub process returned. In the final step the telementry is saved and
closed

Arguments
    - source_name (str):
        Process name for the pipeline process that this telemetry is reporting
        on. Name is free to choose but make sure it is a unique name for each
        unique process as it will otherwise be difficult to use the telemetry
        for monitoring the status of your pipeline processes
    - process_type (str):
        Is one of predefined set of process_type, for example
        `create_data_from_url` or `upload_data`. It is possible to define
        custom process_types. See section on process types.
    - telemetry_rules (dict):
        dict describing any telemetry rules that need to be processed when data
        is added to the telemetry object. With these rules you can define custom
        conditional counts and errors that can be applied to the provided data.
        See telemetry rules section for more detail
    -
"""
from errors import ListErrors

from .decorator import add_mongo_telemetry, add_telemetry
from .helper import increase_base_count, increase_fail_count, \
    is_telemetry_counter
from .main import Telemetry
from .settings.data_class import ProcessType, TelemetryCounter
from .settings.process_type import ProcessTypes
from .settings.settings import BaseEnumerator, DefaultProcessTypes
from .settings.telemetry_errors import ValidationErrors

ProcessTypes.register_process_types(DefaultProcessTypes)
ListErrors.register_errors(ValidationErrors)



"""Module to test storage module.
"""
import sqlite3
from datetime import datetime

from freezegun import freeze_time

from pipeline_telemetry.settings.settings import DEFAULT_TRAFIC_LIGHT_COLOR
from pipeline_telemetry.storage.generic import AbstractTelemetryStorage
from pipeline_telemetry.storage.memory import TelemetryInMemoryStorage
from pipeline_telemetry.storage.mongo import TelemetryMongoModel, \
    TelemetryMongoStorage
from pipeline_telemetry.storage.mongo_connection import get_mongo_db_port


def test_abstract_strorage_class_exists():
    """Test abstract storage class exists."""
    assert AbstractTelemetryStorage


def test_in_memory_strorage_class_exists():
    """Test that in memory storage class exists."""
    assert TelemetryInMemoryStorage


def test_in_memory_strorage_creates_db_in_memory():
    """Test in memory storage instance has db_in_memory and db_cursor"""
    in_memory_storage = TelemetryInMemoryStorage()
    assert in_memory_storage.db_in_memory
    assert in_memory_storage.db_cursor
    assert isinstance(in_memory_storage.db_in_memory, sqlite3.Connection)
    assert isinstance(in_memory_storage.db_cursor, sqlite3.Cursor)


def test_in_memory_strorage_is_only_created_once():
    """Test that once the database is created it will not be reinitialized"""
    in_memory_storage = TelemetryInMemoryStorage()
    db = in_memory_storage.db_in_memory
    TelemetryInMemoryStorage.initialize_db()
    assert in_memory_storage.db_in_memory is db
    new_in_memory_storage = TelemetryInMemoryStorage()
    assert new_in_memory_storage.db_in_memory is db


def test_store_telemetry_stores_object():
    """Test in memory storage instance has db_in_memory and db_cursor"""
    telemetry = {"source_name": "test"}
    # Table reset for each test is needed as the table is a class property
    TelemetryInMemoryStorage._define_db_table(TelemetryInMemoryStorage.db_cursor)
    in_memory_storage = TelemetryInMemoryStorage()
    in_memory_storage.store_telemetry(telemetry)
    all_in_memory_objects = in_memory_storage.db_cursor.execute(
        "SELECT * FROM telemetry "
    )
    assert len(all_in_memory_objects.fetchall()) == 1


@freeze_time("2022-01-18 18:00:00.123456")
def test_store_telemetry_stores_object_with_create_date(mocker):
    """Test in memory storage instance has db_in_memory and db_cursor"""
    telemetry = {"source_name": "test", "start_date_time": datetime.now()}
    # Table reset for each test is needed as the table is a class property
    TelemetryInMemoryStorage._define_db_table(TelemetryInMemoryStorage.db_cursor)
    in_memory_storage = TelemetryInMemoryStorage()
    in_memory_storage.store_telemetry(telemetry)
    in_memory_record = in_memory_storage.db_cursor.execute(
        "SELECT * FROM telemetry LIMIT 1"
    ).fetchone()
    print(in_memory_record)
    assert in_memory_record[4] == "2022-01-18 18:00:00.123456"


def test_telemetry_mongo_model_class_exists():
    """Test TelemetryMongoModel class exists."""
    assert TelemetryMongoModel


def test_telemetry_mongo_storage_class_exists():
    """Test TelemetryMongoStorage class exists."""
    assert TelemetryMongoStorage


def test_store_telemetry_method_processes_and_saves_telemetry(mocker):
    """
    Test store_telemetry method calls _telemetry_model_kwargs and then creates
    and saves a mongo object.
    """
    mongo_module_path = "pipeline_telemetry.storage.mongo."
    mocker.patch(mongo_module_path + "TelemetryMongoModel.save", return_value=True)
    mocker.patch(
        mongo_module_path + "TelemetryMongoStorage._telemetry_model_kwargs",
        return_value={},
    )
    _telemetry_model_kwargs_spy = mocker.spy(
        TelemetryMongoStorage, "_telemetry_model_kwargs"
    )
    _telemetry_mongo_model_new_spy = mocker.spy(TelemetryMongoModel, "__new__")
    _telemetry_mongo_model_save_spy = mocker.spy(TelemetryMongoModel, "save")
    TelemetryMongoStorage().store_telemetry(telemetry={"source_name": "test"})
    assert _telemetry_model_kwargs_spy.called
    assert _telemetry_mongo_model_new_spy.called
    assert _telemetry_mongo_model_save_spy.called


def test_telemetry_model_kwargs_method():
    """Test _telemetry_model_kwargs method returns correct kwargs."""
    result = TelemetryMongoStorage()._telemetry_model_kwargs(
        {
            "category": "test",
            "sub_category": "sub_test",
            "source_name": "tst_source_name",
            "process_type": "tst_process_type",
            "start_date_time": "tst_start_date_time",
            "run_time_in_seconds": "tst_run_time_in_seconds",
            "field1": {"a": 1},
            "field2": "value",
            "traffic_light": DEFAULT_TRAFIC_LIGHT_COLOR,
        }
    )

    assert result == {
        "category": "test",
        "sub_category": "sub_test",
        "source_name": "tst_source_name",
        "process_type": "tst_process_type",
        "start_date_time": "tst_start_date_time",
        "run_time_in_seconds": "tst_run_time_in_seconds",
        "telemetry": {"field1": {"a": 1}, "field2": "value"},
        "traffic_light": DEFAULT_TRAFIC_LIGHT_COLOR,
    }


def test_get_mongo_db_port_returns_none():
    """Test _get_mongo_db_port returns None if no env variable is set."""
    assert get_mongo_db_port() is None


def test_get_mongo_db_port_returns_port_int(mocker):
    """Test _get_mongo_db_port returns port nr as intno env variable is set."""
    mocker.patch("os.getenv", return_value="12345")
    assert get_mongo_db_port() == 12345


def test_default_access_params():
    """Test default_access_params are set poperly."""
    from pipeline_telemetry.storage import mongo_connection as mc

    def_params = mc.MONGO_ACCESS_PARAMS
    assert def_params["port"] == mc.DEFAULT_MONGO_DB_PORT
    assert def_params["host"] == mc.DEFAULT_MONGO_DB_HOST



"""Module to define tests for adding telemetry"""
import pytest
from errors.error import ListErrors
from test_data import DEFAULT_TELEMETRY_PARAMS, TEST_TELEMETRY_RULES

from pipeline_telemetry.main import ERRORS_KEY, Telemetry
from pipeline_telemetry.settings import exceptions


def test_add_method_adds_errors_to_sub_process():
    """Test that errors are added to telemetry sub_process."""
    telemetry = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry.add('RETRIEVE_RAW_DATA', None, [ListErrors.KEY_NOT_FOUND])
    sub_process_telemetry = telemetry.telemetry.get('RETRIEVE_RAW_DATA')
    assert sub_process_telemetry[ERRORS_KEY].get(
        ListErrors.KEY_NOT_FOUND.code) == 1


def test_add_method_raises_exception_when_telemetry_closed():
    """Test that errors are added to telemetry sub_process."""
    telemetry = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry.save_and_close()
    with pytest.raises(exceptions.TelemetryObjectAlreadyClosed):
        telemetry.add('RETRIEVE_RAW_DATA', None, [ListErrors.KEY_NOT_FOUND])


def test_add_errors_method_adds_errors_to_sub_process():
    """Test that errors are added to telemetry sub_process."""
    telemetry = Telemetry(**DEFAULT_TELEMETRY_PARAMS)
    telemetry.increase_sub_process_base_count('RETRIEVE_RAW_DATA')
    telemetry._add_errors('RETRIEVE_RAW_DATA', [ListErrors.KEY_NOT_FOUND])
    sub_process_telemetry = telemetry.telemetry.get('RETRIEVE_RAW_DATA')
    assert sub_process_telemetry[ERRORS_KEY].get(
        ListErrors.KEY_NOT_FOUND.code) == 1


def test_validate_data_method():
    """Test _validate_data_method returns list of errors."""
    telemetry = Telemetry(telemetry_rules=TEST_TELEMETRY_RULES,
                          **DEFAULT_TELEMETRY_PARAMS)
    telemetry.increase_sub_process_base_count('RETRIEVE_RAW_DATA')
    errors = telemetry._validate_data(sub_process='RETRIEVE_RAW_DATA',
                                      data={})
    assert ListErrors.KEY_NOT_FOUND.code in str(errors[0])
    assert '@KEY_<items>' in str(errors[0])


def test_validate_data_method_returns_empty_list():
    """
    Test _validate_data_method returns empty list when no errors are found.
    """
    telemetry = Telemetry(telemetry_rules=TEST_TELEMETRY_RULES,
                          **DEFAULT_TELEMETRY_PARAMS)
    telemetry.increase_sub_process_base_count('RETRIEVE_RAW_DATA')
    errors = telemetry._validate_data(sub_process='RETRIEVE_RAW_DATA',
                                      data={'items': [1, 2, 3]})
    assert errors == []



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

import os
import subprocess
import sys
from os.path import abspath, dirname, exists, join

base_path = dirname(dirname(abspath(__file__)))


def check_call(args):
    print("+", *args)
    subprocess.check_call(args)


def exec_in_env():
    env_path = join(base_path, ".tox", "bootstrap")
    if sys.platform == "win32":
        bin_path = join(env_path, "Scripts")
    else:
        bin_path = join(env_path, "bin")
    if not exists(env_path):
        import subprocess

        print("Making bootstrap env in: {0} ...".format(env_path))
        try:
            check_call([sys.executable, "-m", "venv", env_path])
        except subprocess.CalledProcessError:
            try:
                check_call([sys.executable, "-m", "virtualenv", env_path])
            except subprocess.CalledProcessError:
                check_call(["virtualenv", env_path])
        print("Installing `jinja2` into bootstrap environment...")
        check_call([join(bin_path, "pip"), "install", "jinja2", "tox"])
    python_executable = join(bin_path, "python")
    if not os.path.exists(python_executable):
        python_executable += '.exe'

    print("Re-executing with: {0}".format(python_executable))
    print("+ exec", python_executable, __file__, "--no-env")
    os.execv(python_executable, [python_executable, __file__, "--no-env"])


def main():
    import jinja2

    print("Project path: {0}".format(base_path))

    jinja = jinja2.Environment(
        loader=jinja2.FileSystemLoader(join(base_path, "ci", "templates")),
        trim_blocks=True,
        lstrip_blocks=True,
        keep_trailing_newline=True,
        autoescape=True,
    )

    tox_environments = [
        line.strip()
        # 'tox' need not be installed globally, but must be importable
        # by the Python that is running this script.
        # This uses sys.executable the same way that the call in
        # cookiecutter-pylibrary/hooks/post_gen_project.py
        # invokes this bootstrap.py itself.
        for line in subprocess.check_output([sys.executable, '-m', 'tox', '--listenvs'], universal_newlines=True).splitlines()
    ]
    tox_environments = [line for line in tox_environments if line.startswith('py')]

    for name in os.listdir(join("ci", "templates")):
        with open(join(base_path, name), "w") as fh:
            fh.write(jinja.get_template(name).render(tox_environments=tox_environments))
        print("Wrote {}".format(name))
    print("DONE.")


if __name__ == "__main__":
    args = sys.argv[1:]
    if args == ["--no-env"]:
        main()
    elif not args:
        exec_in_env()
    else:
        print("Unexpected arguments {0}".format(args), file=sys.stderr)
        sys.exit(1)



"""
module to test validate module of data-validator
"""
import pytest
from test_data import InstructionTestClass

from pipeline_telemetry.settings import exceptions
from pipeline_telemetry.validators.dict_validator import DictValidator
from pipeline_telemetry.validators.has_key import HasKey

# pylint: disable=protected-access


def test_dict_validator_class_exists():
    """ check that DictValidator class exists """
    assert DictValidator


def test_dict_validator_has_instructions():
    """ check that DictValidator class has instructions attribute"""
    assert DictValidator._instructions


def test_dict_validator_register_instruction():
    """
    check that register_instruction method registers an instruction
    """
    # ensure that test is not corrupted by previous registrations
    DictValidator._instructions = {}
    DictValidator.register_instruction(InstructionTestClass)
    assert InstructionTestClass.instruction in \
        list(DictValidator._instructions.keys())


def test_dict_validator_register_instruction_twice():
    """
    check that register_instruction method raises exception if you try to
    register an instruction twice
    """
    # ensure that test is not corrupted by previous registrations
    DictValidator._instructions = {}
    DictValidator.register_instruction(InstructionTestClass)
    with pytest.raises(exceptions.InstructionRegisteredTwice):
        DictValidator.register_instruction(InstructionTestClass)


def test_apply_rule_raises_exception_with_unregistered_instruction():
    """
    test that _apply_rule method returns an exception when a rule with an
    unregistered instruction is provided
    """
    with pytest.raises(exceptions.UnknownInstruction):
        DictValidator()._apply_rule(
            {}, ('unkwown_rule', {'rule details': 'details'}))


def test_apply_rule_calls_do_validate_method(mocker):
    """
    test that _apply_rule method calls the `validate` method on the instruction
    class and returns the value receive from the validate method
    """
    # as previous tests might have cleaned the instruction set it needs to be
    # reregistered
    DictValidator.register_instruction(HasKey)
    rule = ('has_key', {'has_key_details': 'rule'})
    mocker.patch('pipeline_telemetry.validators.has_key.HasKey.validate',
                 return_value='return_value')
    _do_validate_spy = mocker.spy(
        HasKey, 'validate')
    return_value = DictValidator()._apply_rule({}, rule)
    assert return_value == 'return_value'
    assert _do_validate_spy.called


def test_instruction_from_rule_succes():
    """ test instruction returns instruction when given a valid rule """
    valid_rule = ('instruction_in_valid_rule',  {'rule_details': 'test'})
    assert DictValidator._instruction_from_rule(valid_rule) == \
        'instruction_in_valid_rule'


def test_instruction_from_rule_with_more_the_one_rule():
    """ test instruction returns instruction when given a valid rule """
    to_many_rules = {'rule_1': 'test1', 'rule_2': 'test2'}
    with pytest.raises(exceptions.RuleCanHaveOnlyOneInstruction):
        DictValidator._instruction_from_rule(to_many_rules)



"""Module to define abstract storage class
"""
from abc import ABCMeta, abstractmethod


class AbstractTelemetryStorage(metaclass=ABCMeta):
    """ Abstract Telemetry Storage class

    implements store_telemetry method that persists a given telemetry object

    Any class that stores the telemetry objects should be subclassed from
    this Abstract Class
    """
    # pylint: disable=too-few-public-methods

    @abstractmethod
    def store_telemetry(self, telemetry: dict) -> None:
        """ public method to persist telemetry object"""



"""Module to define validate entries validator class
"""
import jmespath
from errors import ErrorCode, ListErrors, add_error_data

from ..settings import exceptions
from ..validators.dict_validator import DictValidator


class ValidateEntries():
    """
    class to define validate entries instruction.
    This instruction counts the entries in a list or a dict and reports an
    error when the number of items in the list or dict is different from the
    expected nr of items
    """
    instruction = 'validate_entries'
    fieldname_field = 'field_name'
    expected_count_field = 'expected_count'

    def __new__(cls):
        """ make this a singleton class """
        return cls

    @classmethod
    def validate(
            cls, dict_to_validate: dict, rule_content: dict) -> list[ErrorCode]:
        """
        Public method run the validation
        prior to validation the rule will be validated.

        Args:
            dict_to_validate (dict): Input dict that needs validation
            rule_content (dict): actual rule

        Returns:
            list[ErrorCode]: List with errors that were found

        Raise:
            FieldNameMandatory: when field name not defined in rule
        """
        cls._validate_rule(rule_content)
        return cls._validate(dict_to_validate, rule_content)

    @classmethod
    def _validate(
            cls, dict_to_validate: dict,
            rule_content: dict) -> list[ErrorCode]:
        """
        method to do the actual validation

        returns:
            - error in case
        """
        fieldname = cls._get_field_name(rule_content)

        field_to_validate = jmespath.search(fieldname, dict_to_validate)

        if not field_to_validate:
            return cls._validation_error(
                ListErrors.FIELD_NOT_FOUND, fieldname)

        if not isinstance(field_to_validate, (list, dict)):
            return cls._validation_error(
                ListErrors.WRONG_TYPE_IN_FIELD, fieldname)

        if len(field_to_validate) != cls._get_expected_count(rule_content):
            return cls._validation_error(
                ListErrors.UNEXPECTED_NR_OF_ITEMS, fieldname)

        return []

    @classmethod
    def _validate_rule(cls, rule_content):
        fieldname = cls._get_field_name(rule_content)
        if not isinstance(fieldname, str):
            raise exceptions.FieldNameMandatory(cls.instruction)
        expected_count = cls._get_expected_count(rule_content)
        if not isinstance(expected_count, int) or expected_count < 0:
            raise exceptions.ExpectedCountMustBePositiveInt()

    @classmethod
    def _get_field_name(cls, rule_content):
        """
        retrieves the field name in scope from the rule content
        """
        return rule_content.get(cls.fieldname_field, None)

    @classmethod
    def _get_expected_count(cls, rule_content):
        """
        retrieves the expected_count in scope of the rule content
        """
        return rule_content.get(cls.expected_count_field, None)

    @staticmethod
    def _validation_error(
            error_code: ErrorCode, fieldname: str) -> list[ErrorCode]:
        """ returns error code object in list with fieldname as error data """
        return [add_error_data(error=error_code, error_data=fieldname)]


DictValidator.register_instruction(ValidateEntries)



"""Module to provide a storage class for using MongoDB.

When using MongoDB storage class you can select the mongoDB instance to be used
via environment Variables

MONGO_DB_NAME (defaults to `telemetry`)
MONGO_DB_PASSWORD
MONGO_DB_USERNAME
MONGO_DB_HOST
MONGO_DB_PORT

If no host and port are defined the connection will dedault to a localhost
mongoDB instance.
"""
from mongoengine import DateTimeField, DictField, Document, StringField, \
    connect

from .generic import AbstractTelemetryStorage
from .mongo_connection import MONGO_ACCESS_PARAMS

connect(alias="telemetry", **MONGO_ACCESS_PARAMS)


class TelemetryMongoModel(Document):
    """
    Class to provice telemetry Mongo Model for persistance in MongoDB
    """

    category = StringField()
    sub_category = StringField()
    source_name = StringField()
    process_type = StringField()
    start_date_time = DateTimeField()
    run_time_in_seconds = StringField()
    traffic_light = StringField()
    telemetry = DictField(default=None)

    meta = {
        "db_alias": "telemetry",
        "indexes": [
            "category",
            "sub_category",
            "source_name",
            ("category", "sub_category", "source_name", "process_type"),
            "process_type",
            "traffic_light",
            "start_date_time",
        ],
    }


class TelemetryMongoStorage(AbstractTelemetryStorage):
    """
    Class to provice telemetry in mongo storage class.
    This class can be used as storage_class argument when creating
    an instance of Telemetry.
    """

    def store_telemetry(self, telemetry: dict) -> None:
        """public method to persist telemetry object"""
        telemetry_mongo_kwargs = self._telemetry_model_kwargs(telemetry)
        TelemetryMongoModel(**telemetry_mongo_kwargs).save()

    @staticmethod
    def _telemetry_model_kwargs(telemetry: dict) -> dict:
        """
        Returns a dicts with kwargs that can be used to create a new
        TelemetryMongoStorage instance.
        """
        telemetry_copy = telemetry.copy()
        category = telemetry_copy.pop("category", None)
        sub_category = telemetry_copy.pop("sub_category", None)
        source_name = telemetry_copy.pop("source_name", None)
        process_type = telemetry_copy.pop("process_type", None)
        start_date_time = telemetry_copy.pop("start_date_time", None)
        traffic_light = telemetry_copy.pop("traffic_light", None)
        run_time_in_seconds = telemetry_copy.pop("run_time_in_seconds", None)
        return {
            "category": category,
            "sub_category": sub_category,
            "source_name": source_name,
            "process_type": process_type,
            "start_date_time": start_date_time,
            "run_time_in_seconds": run_time_in_seconds,
            "traffic_light": traffic_light,
            "telemetry": telemetry_copy,
        }



"""Module to test decorator logic.
"""
from test_data import DEFAULT_TELEMETRY_PARAMS

from pipeline_telemetry.decorator import add_mongo_telemetry, add_telemetry
from pipeline_telemetry.storage.memory import TelemetryInMemoryStorage
from pipeline_telemetry.storage.mongo import TelemetryMongoStorage


def test_default_decorator():
    """
    Test that default decorator does not effect decorated method result
    but sets the telemetry property.
    """

    class DecoratorTest:
        @add_telemetry(DEFAULT_TELEMETRY_PARAMS)
        def decorated_method(self):
            return "method result"

    class_instance = DecoratorTest()
    assert not hasattr(class_instance, "_telemetry")
    assert class_instance.decorated_method() == "method result"
    assert hasattr(class_instance, "_telemetry")


def test_mongo_telemetry_decorator(mocker):
    """
    Test that mongo decorator does not effect decorated method result
    but sets the telemetry property.
    """
    mocker.patch("pipeline_telemetry.decorator.TelemetryMongoStorage",
                 TelemetryInMemoryStorage)

    class DecoratorTest:
        @add_mongo_telemetry(DEFAULT_TELEMETRY_PARAMS)
        def decorated_method(self):
            return "method result"

    class_instance = DecoratorTest()
    assert not hasattr(class_instance, "_telemetry")
    assert class_instance.decorated_method() == "method result"
    assert hasattr(class_instance, "_telemetry")


def test_mongo_telemetry_decorator_sets_mongo_storage_class(mocker):
    """
    Test that mongo decorator sets the mongo storage class
    """
    mongo_module_path = "pipeline_telemetry.storage.mongo."
    mocker.patch(mongo_module_path + "TelemetryMongoModel.save",
                 return_value=None)

    class DecoratorTest:
        @add_mongo_telemetry(DEFAULT_TELEMETRY_PARAMS)
        def decorated_method(self):
            return self._telemetry.storage_class

    assert isinstance(DecoratorTest().decorated_method(), TelemetryMongoStorage)


def test_calling_decorated_method_from_within_decorated_method(mocker):
    """
    Test that when telemetry is active it is not changed by any other telemetry
    decorated method that is being called.
    """
    mongo_module_path = "pipeline_telemetry.storage.mongo."
    mocker.patch(mongo_module_path + "TelemetryMongoModel.save",
                 return_value=None)

    changed_telemetry_params = DEFAULT_TELEMETRY_PARAMS.copy() | {
        "category": "OTHER VALUE",
    }

    class DecoratorTest:
        @add_mongo_telemetry(DEFAULT_TELEMETRY_PARAMS)
        def decorated_method(self):
            return self.second_decorated_method()

        @add_mongo_telemetry(changed_telemetry_params)
        def second_decorated_method(self):
            return self._telemetry.get("category")

    class_instance = DecoratorTest()
    assert not hasattr(class_instance, "_telemetry")
    assert class_instance.decorated_method() == \
        DEFAULT_TELEMETRY_PARAMS.get("category")
    assert hasattr(class_instance, "_telemetry")


def test_calling_mongo_decorated_method_from_within_decorated_method():
    """
    Test that when a telemetry instance created with add_mongo_telemetry is
    active it is not changed by any other telemetry decorated method that is
    being called.
    """
    changed_telemetry_params = DEFAULT_TELEMETRY_PARAMS.copy() | {
        "category": "OTHER VALUE",
    }

    class DecoratorTest:
        @add_telemetry(DEFAULT_TELEMETRY_PARAMS)
        def decorated_method(self):
            return self.second_decorated_method()

        @add_telemetry(changed_telemetry_params)
        def second_decorated_method(self):
            return self._telemetry.get("category")

    class_instance = DecoratorTest()
    assert not hasattr(class_instance, "_telemetry")
    assert class_instance.decorated_method() == DEFAULT_TELEMETRY_PARAMS.get("category")
    assert hasattr(class_instance, "_telemetry")



"""
Module to define the DictValidator class

classes:
    - DictValidator
"""
from typing import List

from errors import ErrorCode

from ..settings.exceptions import InstructionRegisteredTwice, \
    RuleCanHaveOnlyOneInstruction, UnknownInstruction


class DictValidator():
    """Class to allow a dict validation according to a set of rules.

    class methods:
        - register_instruction():
            method to add instruction sets to the dictvalidator
    public methods:
        - validate():
            method to run the validate of a dict
    """

    _instructions = {}

    @classmethod
    def validate(cls,
                 dict_to_validate: dict,
                 validation_rules: dict) -> List[ErrorCode]:
        """Public class method to run the validation.

        :param dict_to_validate: data dict to be validated
        :type dict_to_validate: dict
        :param validation_rules: validation rules for this data object
        :type validation_rules: dict

        :returns: List of ErrorCodes. List is empty if no Errors are found
        """
        errors = []
        for rule in validation_rules.items():
            errors.extend(cls._apply_rule(dict_to_validate, rule))
        return errors

    @classmethod
    def _apply_rule(
            cls, dict_to_validate: dict, rule: tuple) -> List[ErrorCode]:
        """[summary]

        Args:
            dict_to_validate (dict): data dict to be validated
            rule (tuple): The rule with the instruction on position 0 and
                          details for the instruction on position 1

        Raises:
            UnknownInstruction: [description]

        Returns:
            list: [description]
        """
        instruction = cls._instruction_from_rule(rule)
        if instruction not in cls._instructions:
            raise UnknownInstruction(instruction)
        return cls._instructions.get(instruction).validate(
            dict_to_validate=dict_to_validate,
            rule_content=rule[1]
        )

    @classmethod
    def register_instruction(cls, instruction_class: type) -> None:
        """Registration at class level of the instruction

        Args:
            instruction_class (type):

        Raises:
            InstructionRegisteredTwice: [description]
        """
        instruction = instruction_class.instruction
        if instruction in cls._instructions:
            raise InstructionRegisteredTwice(instruction_class)
        cls._instructions.update({
            instruction_class.instruction: instruction_class})

    @staticmethod
    def _instruction_from_rule(rule: tuple) -> str:
        """Returns the instuction string from rule dict

        Args:
            rule (tuple): The rule with the instruction on position 0 and
                          details for the instruction on position 1

        Raises:
            RuleCanHaveOnlyOneInstruction:
                if 0 or multiple instructions were given

        Returns:
            str: the instruction
        """
        if not (isinstance(rule, tuple) and len(rule) == 2):
            raise RuleCanHaveOnlyOneInstruction(rule)
        return rule[0]



"""Module to run integration test for telemetry module
"""
from test_data import DEFAULT_TELEMETRY_PARAMS
from test_integration_data import BASIC_RETRIEVE_DATA_TEST, \
    ITEMS_FIELD_HAS_DICT_TEST, MISSING_KEY_RETRIEVE_DATA_TEST, \
    MUST_HAVE_KEY_IN_ITEMS_TEST, NO_ITEMS_HAVE_MUST_HAVE_KEY_TEST, \
    validate_result_from_telemetry

from pipeline_telemetry.main import Telemetry


def test_basic_test_data():
    """Test data basic data test returns correct telemetry."""
    telemetry_rules = BASIC_RETRIEVE_DATA_TEST.validation_rules
    telemetry_params = DEFAULT_TELEMETRY_PARAMS | \
        {'telemetry_rules': telemetry_rules}
    telemetry = Telemetry(**telemetry_params)
    telemetry.increase_sub_process_base_count('RETRIEVE_RAW_DATA')
    telemetry.add('RETRIEVE_RAW_DATA', BASIC_RETRIEVE_DATA_TEST.data, None)
    telemetry.save_and_close()
    assert validate_result_from_telemetry(telemetry, BASIC_RETRIEVE_DATA_TEST)


def test_missing_key_integration_test():
    """Test telemetry returns correct telemetry object when key is missing."""
    telemetry_rules = MISSING_KEY_RETRIEVE_DATA_TEST.validation_rules
    telemetry_params = DEFAULT_TELEMETRY_PARAMS | \
        {'telemetry_rules': telemetry_rules}
    telemetry = Telemetry(**telemetry_params)
    telemetry.increase_sub_process_base_count('RETRIEVE_RAW_DATA')
    telemetry.add('RETRIEVE_RAW_DATA',
                  MISSING_KEY_RETRIEVE_DATA_TEST.data, None)
    telemetry.save_and_close()
    assert validate_result_from_telemetry(
        telemetry, MISSING_KEY_RETRIEVE_DATA_TEST)


def test_must_have_key_in_items_integration_test():
    """
    Test telemetry returns correct telemetry object when key is missing in one
    of the items.
    """
    telemetry_rules = MUST_HAVE_KEY_IN_ITEMS_TEST.validation_rules
    telemetry_params = DEFAULT_TELEMETRY_PARAMS | \
        {'telemetry_rules': telemetry_rules}
    telemetry = Telemetry(**telemetry_params)
    telemetry.increase_sub_process_base_count('RETRIEVE_RAW_DATA')
    telemetry.add('RETRIEVE_RAW_DATA',
                  MUST_HAVE_KEY_IN_ITEMS_TEST.data, None)
    telemetry.save_and_close()
    assert validate_result_from_telemetry(
        telemetry, MUST_HAVE_KEY_IN_ITEMS_TEST)


def test_no_items_have_must_have_key_in_items_integration_test():
    """
    Test telemetry returns correct telemetry object when no item has the
    correct key.
    """
    telemetry_rules = NO_ITEMS_HAVE_MUST_HAVE_KEY_TEST.validation_rules
    telemetry_params = DEFAULT_TELEMETRY_PARAMS | \
        {'telemetry_rules': telemetry_rules}
    telemetry = Telemetry(**telemetry_params)
    telemetry.increase_sub_process_base_count('RETRIEVE_RAW_DATA')
    telemetry.add('RETRIEVE_RAW_DATA',
                  NO_ITEMS_HAVE_MUST_HAVE_KEY_TEST.data, None)
    telemetry.save_and_close()
    assert validate_result_from_telemetry(
        telemetry, NO_ITEMS_HAVE_MUST_HAVE_KEY_TEST)


def test_items_has_dict_field_test():
    """
    Test telemetry returns correct telemetry object when items field contains
    a dict field.
    """
    telemetry_rules = ITEMS_FIELD_HAS_DICT_TEST.validation_rules
    telemetry_params = DEFAULT_TELEMETRY_PARAMS | \
        {'telemetry_rules': telemetry_rules}
    telemetry = Telemetry(**telemetry_params)
    telemetry.increase_sub_process_base_count('RETRIEVE_RAW_DATA')
    telemetry.add('RETRIEVE_RAW_DATA',
                  ITEMS_FIELD_HAS_DICT_TEST.data, None)
    telemetry.save_and_close()
    print(telemetry.telemetry)
    assert validate_result_from_telemetry(
        telemetry, ITEMS_FIELD_HAS_DICT_TEST)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import io
import re
from glob import glob
from os.path import basename, dirname, join, splitext

from setuptools import find_packages, setup


def read(*names, **kwargs):
    with io.open(
        join(dirname(__file__), *names), encoding=kwargs.get("encoding", "utf8")
    ) as fh:
        return fh.read()


setup(
    name="pipeline-telemetry",
    version="0.2.15",
    license="LGPL-3.0-or-later",
    description="Create and store data pipeline telemetry data",
    long_description="%s\n%s"
    % (
        re.compile("^.. start-badges.*^.. end-badges", re.M | re.S).sub(
            "", read("README.rst")
        ),
        re.sub(":[a-z]+:`~?(.*?)`", r"``\1``", read("CHANGELOG.rst")),
    ),
    author="Maarten de Ruyter",
    author_email="maarten@geodatagarden.com",
    url="https://github.com/MaartendeRuyter/pipeline-telemetry",
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.9",
        # 'Programming Language :: Python :: Implementation :: CPython',
        # 'Programming Language :: Python :: Implementation :: PyPy',
        # # uncomment if you test on these interpreters:
        # 'Programming Language :: Python :: Implementation :: IronPython',
        # 'Programming Language :: Python :: Implementation :: Jython',
        # 'Programming Language :: Python :: Implementation :: Stackless',
        "Topic :: Utilities",
    ],
    project_urls={
        "Documentation": "https://pipeline-telemetry.readthedocs.io/",
        "Changelog": "https://pipeline-telemetry.readthedocs.io/en/latest/changelog.html",
        "Issue Tracker": "https://github.com/MaartendeRuyter/pipeline-telemetry/issues",
    },
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    python_requires=">=3.9",
    install_requires=[
        "jmespath>=0.10.0",
        "error-manager>=1.1.1"
        # eg: 'aspectlib==1.1.1', 'six>=1.7',
    ],
    extras_require={
        # eg:
        #   'rst': ['docutils>=0.11'],
        #   ':python_version=="2.6"': ['argparse'],
    },
    entry_points={
        "console_scripts": [
            "pipeline-telemetry = pipeline_telemetry.cli:main",
        ]
    },
)



"""Helper module for pipeline_telemetry
"""
from typing import Any, Union

from pipeline_telemetry.settings.data_class import TelemetryCounter


def is_telemetry_counter(counter: Union[TelemetryCounter, Any]) -> bool:
    """Method to check in object is an instance of TelemetryCounter."""
    return issubclass(counter.__class__, TelemetryCounter)


def increase_base_count(
    object_with_telemetry: Any, sub_process: str, increment: int = 1
) -> None:
    """
    Helper method to increase base count for a sub_process
    """
    object_with_telemetry._telemetry.increase_sub_process_base_count(
        sub_process=sub_process, increment=increment
    )


def increase_fail_count(
    object_with_telemetry: Any, sub_process: str, increment: int = 1
) -> None:
    """
    Helper method to increase base count for a sub_process
    """
    object_with_telemetry._telemetry.increase_sub_process_fail_count(
        sub_process=sub_process, increment=increment
    )



"""
module to test has_key instruction class
"""
# pylint: disable=protected-access
import pytest
from test_data import TEST_ERROR_CODE

from pipeline_telemetry.settings.exceptions import \
    ExpectedCountMustBePositiveInt, FieldNameMandatory
from pipeline_telemetry.settings.telemetry_errors import ErrorCode
from pipeline_telemetry.validators.validate_entries import ValidateEntries


def test_validate_entries_class_exists():
    """ check that ValidateEntries class exists """
    assert ValidateEntries


def test_validate_entries_class_is_singleton_clss():
    """Check that ValidateEntries class is a singleton class."""
    assert ValidateEntries() is ValidateEntries()


def test_validate_rule_method_raises_exception():
    """
    check that _validate_rule method raises exception when field name
    not provided in rule
    """
    rule_content = {'not_field_name': 'test', 'expected_count': 1}
    with pytest.raises(FieldNameMandatory) as exception:
        ValidateEntries._validate_rule(rule_content)

    assert 'For `validate_entries` instruction' in str(exception)


def test_validate_rule_method_without_expected_count_raises_exception():
    """
    check that _validate_rule method raises exception when expected_count
    not provided in rule
    """
    rule_content = {'field_name': 'test'}
    with pytest.raises(ExpectedCountMustBePositiveInt) as exception:
        ValidateEntries._validate_rule(rule_content)

    assert 'Ruleset does not contain `expected_count`' in str(exception)


def test_validate_rule_method_with_neg_expected_count_raises_exception():
    """
    check that _validate_rule method raises exception when expected_count
    is a negative integer
    """
    rule_content = {'field_name': 'test', 'expected_count': -1}
    with pytest.raises(ExpectedCountMustBePositiveInt):
        ValidateEntries._validate_rule(rule_content)


def test_validate_rule_method_returns_none():
    """
    check that _validate_rule method returns none when field name and
    expected_count are provided in the rule
    """
    rule_content = {'field_name': 'test', 'expected_count': 1}
    assert ValidateEntries._validate_rule(rule_content) is None


def test_get_field_name_method():
    """
    check that _get_field_name method returns value of
    field_name that is in scope of validation
    """
    rule_content = {'field_name': 'test', 'expected_count': 1}
    assert ValidateEntries._get_field_name(rule_content) == 'test'


def test_get_expected_count():
    """
    check that _get_expected_count method returns value of
    expected_count
    """
    rule_content = {'field_name': 'test', 'expected_count': 1}
    assert ValidateEntries._get_expected_count(rule_content) == 1


def test_validation_error_method():
    """
    check that _validation_error method returns instance
    of ErrorCode with fieldname in error_data
    """
    validation_error_list = ValidateEntries._validation_error(
        TEST_ERROR_CODE, 'test_error_data')
    validation_error = validation_error_list[0]
    assert isinstance(validation_error_list, list)
    assert isinstance(validation_error, ErrorCode)
    assert validation_error.error_data == 'test_error_data'


def test_validate_valid_dict():
    """
    Test that a dict with a field with a list with the expected count (length)
    returns empty error list
    """
    valid_dict = {'correct_key': [1, 2, 3, 4]}
    rule_content = {'field_name': 'correct_key', 'expected_count': 4}
    validation_result = ValidateEntries.validate(
        dict_to_validate=valid_dict, rule_content=rule_content
    )
    assert validation_result == []


def test_validate_valid_dict_with_nested_key():
    """
    Test that a dict of the right length in nested key returns an empty error
    lits
    """
    valid_dict = {'key': {'nested_key': {'a': 1, 'b': 2}}}
    rule_content = {'field_name': 'key.nested_key', 'expected_count': 2}
    validation_result = ValidateEntries.validate(
        dict_to_validate=valid_dict, rule_content=rule_content
    )
    assert validation_result == []


def test_validate_invalid_dict():
    """
    Test that a dict without right key returns a list with error_code
    VALIDATE_ENTRIES_ERR_001 and the field_name in the error_data

    """
    invalid_dict = {'incorrect_key': 'value'}
    rule_content = {'field_name': 'correct_key', 'expected_count': 2}
    validation_result = ValidateEntries.validate(
        dict_to_validate=invalid_dict, rule_content=rule_content
    )
    assert len(validation_result) == 1
    assert validation_result[0].code == 'VALIDATE_ENTRIES_ERR_001'
    assert validation_result[0].error_data == 'correct_key'


def test_validate_dict_with_wrong_type_in_field():
    """
    Test that a dict without right key returns a list with error_code
    VALIDATE_ENTRIES_ERR_002 and the field_name in the error_data

    """
    invalid_dict = {'correct_key': 'value'}
    rule_content = {'field_name': 'correct_key', 'expected_count': 2}
    validation_result = ValidateEntries.validate(
        dict_to_validate=invalid_dict, rule_content=rule_content
    )
    assert len(validation_result) == 1
    assert validation_result[0].code == 'VALIDATE_ENTRIES_ERR_002'
    assert validation_result[0].error_data == 'correct_key'


def test_validate_dict_with_wrong_number_of_entries():
    """
    Test that a dict without right key returns a list with error_code
    VALIDATE_ENTRIES_ERR_003 and the field_name in the error_data

    """
    invalid_dict = {'correct_key': [1, 2, 3]}
    rule_content = {'field_name': 'correct_key', 'expected_count': 2}
    validation_result = ValidateEntries.validate(
        dict_to_validate=invalid_dict, rule_content=rule_content
    )
    assert len(validation_result) == 1
    assert validation_result[0].code == 'VALIDATE_ENTRIES_ERR_003'
    assert validation_result[0].error_data == 'correct_key'



"""
module to test has_key instruction class
"""
import pytest
from test_data import TEST_ERROR_CODE

from pipeline_telemetry.settings.exceptions import FieldNameMandatory
from pipeline_telemetry.settings.telemetry_errors import ErrorCode
from pipeline_telemetry.validators.has_key import HasKey

# pylint: disable=protected-access


def test_has_key_class_exists():
    """ check that HasKey class exists """
    assert HasKey


def test_has_key_class_is_singelton():
    """ check that HasKey is a singleton class """
    instance_one = HasKey()
    instance_two = HasKey()
    assert instance_one is instance_two


def test_validate_rule_method_raises_exception():
    """
    check that _validate_rule method raises exception when field name
    not provided in rule
    """
    rule_content = {'not_field_name': 'test'}
    with pytest.raises(FieldNameMandatory) as exception:
        HasKey._validate_rule(rule_content)

    assert 'For `has_key` instruction' in str(exception)


def test_validate_rule_method_returns_none():
    """
    check that _validate_rule method returns none when field name
    provided in rule
    """
    rule_content = {'field_name': 'test'}
    assert HasKey._validate_rule(rule_content) is None


def test_get_field_name_method():
    """
    check that _get_field_name method returns value of
    field_name that is in scope of validation
    """
    rule_content = {'field_name': 'test'}
    assert HasKey._get_field_name(rule_content) == 'test'


def test_validation_error_method():
    """
    Check that _validation_error method returns a list with
    ErrorCode instance.
    """
    validation_error_list = HasKey._validation_error(
        TEST_ERROR_CODE, 'test_error_data')
    validation_error = validation_error_list[0]
    assert isinstance(validation_error_list, list)
    assert isinstance(validation_error, ErrorCode)


def test_validate_valid_dict():
    """Test that a dict with right key returns empty error list
    """
    valid_dict = {'correct_key': 'value'}
    rule_content = {'field_name': 'correct_key'}
    validation_result = HasKey.validate(
        dict_to_validate=valid_dict, rule_content=rule_content
    )
    assert validation_result == []


def test_validate_valid_dict_with_nested_key():
    """
    Test that a dict with right key returns empty error list when the nested
    key in input dict is found
    """
    valid_dict = {'correct_key': {'nested_key': 'value'}}
    rule_content = {'field_name': 'correct_key.nested_key'}
    validation_result = HasKey.validate(
        dict_to_validate=valid_dict, rule_content=rule_content
    )
    assert validation_result == []


def test_validate_invalid_dict():
    """
    Test that a dict without right key returns a list with error_code
    KEY_NOT_FOUND_001 and the field_name in the error_data

    """
    invalid_dict = {'incorrect_key': 'value'}
    rule_content = {'field_name': 'correct_key'}
    validation_result = HasKey.validate(
        dict_to_validate=invalid_dict, rule_content=rule_content
    )
    assert len(validation_result) == 1
    assert validation_result[0].code == 'HAS_KEY_ERR_0001@KEY_<correct_key>'


def test_validate_invalid_nested_dict():
    """
    Test that a nested dict without right nested key returns a list with
    error_code KEY_NOT_FOUND_001 and the nested field_name in the error_data.
    """
    invalid_dict = {'incorrect_key': {'nested_key': 'value'}}
    rule_content = {'field_name': 'correct_key.nested_key'}
    validation_result = HasKey.validate(
        dict_to_validate=invalid_dict, rule_content=rule_content
    )
    assert len(validation_result) == 1
    assert validation_result[0].code == \
        'HAS_KEY_ERR_0001@KEY_<correct_key.nested_key>'
